# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import copy

import testtools
from testtools.matchers import HasLength

from ironicclient import exc
from ironicclient.tests.unit import utils
import ironicclient.v1.deploy_template

DEPLOY_TEMPLATE = {'uuid': '11111111-2222-3333-4444-555555555555',
                   'name': 'fake-template',
                   'steps': {},
                   'extra': {}}

DEPLOY_TEMPLATE2 = {'uuid': '55555555-4444-3333-2222-111111111111',
                    'name': 'fake-template2',
                    'steps': {},
                    'extra': {}}

CREATE_DEPLOY_TEMPLATE = copy.deepcopy(DEPLOY_TEMPLATE)
del CREATE_DEPLOY_TEMPLATE['uuid']

CREATE_DEPLOY_TEMPLATE_WITH_UUID = copy.deepcopy(DEPLOY_TEMPLATE)

UPDATED_DEPLOY_TEMPLATE = copy.deepcopy(DEPLOY_TEMPLATE)
NEW_NAME = 'fake-template3'
UPDATED_DEPLOY_TEMPLATE['name'] = NEW_NAME

fake_responses = {
    '/v1/deploy_templates':
    {
        'GET': (
            {},
            {"deploy_templates": [DEPLOY_TEMPLATE]},
        ),
        'POST': (
            {},
            CREATE_DEPLOY_TEMPLATE,
        ),
    },
    '/v1/deploy_templates/?detail=True':
    {
        'GET': (
            {},
            {"deploy_templates": [DEPLOY_TEMPLATE]},
        ),
    },
    '/v1/deploy_templates/?fields=uuid,name':
    {
        'GET': (
            {},
            {"deploy_templates": [DEPLOY_TEMPLATE]},
        ),
    },
    '/v1/deploy_templates/%s' % DEPLOY_TEMPLATE['uuid']:
    {
        'GET': (
            {},
            DEPLOY_TEMPLATE,
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            UPDATED_DEPLOY_TEMPLATE,
        ),
    },
    '/v1/deploy_templates/%s?fields=uuid,name' % DEPLOY_TEMPLATE['uuid']:
    {
        'GET': (
            {},
            DEPLOY_TEMPLATE,
        ),
    },
}

fake_responses_pagination = {
    '/v1/deploy_templates':
    {
        'GET': (
            {},
            {"deploy_templates": [DEPLOY_TEMPLATE],
             "next": "http://127.0.0.1:6385/v1/deploy_templates/?limit=1"}
        ),
    },
    '/v1/deploy_templates/?limit=1':
    {
        'GET': (
            {},
            {"deploy_templates": [DEPLOY_TEMPLATE2]}
        ),
    },
    '/v1/deploy_templates/?marker=%s' % DEPLOY_TEMPLATE['uuid']:
    {
        'GET': (
            {},
            {"deploy_templates": [DEPLOY_TEMPLATE2]}
        ),
    },
}

fake_responses_sorting = {
    '/v1/deploy_templates/?sort_key=updated_at':
    {
        'GET': (
            {},
            {"deploy_templates": [DEPLOY_TEMPLATE2, DEPLOY_TEMPLATE]}
        ),
    },
    '/v1/deploy_templates/?sort_dir=desc':
    {
        'GET': (
            {},
            {"deploy_templates": [DEPLOY_TEMPLATE2, DEPLOY_TEMPLATE]}
        ),
    },
}


class DeployTemplateManagerTest(testtools.TestCase):

    def setUp(self) -> None:
        super(DeployTemplateManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = ironicclient.v1.deploy_template.DeployTemplateManager(
            self.api)

    def test_deploy_templates_list(self) -> None:
        deploy_templates = self.mgr.list()
        expect = [
            ('GET', '/v1/deploy_templates', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(deploy_templates))

    def test_deploy_templates_list_detail(self) -> None:
        deploy_templates = self.mgr.list(detail=True)
        expect = [
            ('GET', '/v1/deploy_templates/?detail=True', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(deploy_templates))

    def test_deploy_template_list_fields(self) -> None:
        deploy_templates = self.mgr.list(fields=['uuid', 'name'])
        expect = [
            ('GET', '/v1/deploy_templates/?fields=uuid,name', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(deploy_templates))

    def test_deploy_template_list_detail_and_fields_fail(self) -> None:
        self.assertRaises(exc.InvalidAttribute, self.mgr.list,
                          detail=True, fields=['uuid', 'name'])

    def test_deploy_templates_list_limit(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.deploy_template.DeployTemplateManager(
            self.api)
        deploy_templates = self.mgr.list(limit=1)
        expect = [
            ('GET', '/v1/deploy_templates/?limit=1', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(deploy_templates, HasLength(1))

    def test_deploy_templates_list_marker(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.deploy_template.DeployTemplateManager(
            self.api)
        deploy_templates = self.mgr.list(marker=DEPLOY_TEMPLATE['uuid'])
        expect = [
            ('GET',
             '/v1/deploy_templates/?marker=%s' % DEPLOY_TEMPLATE['uuid'], {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(deploy_templates, HasLength(1))

    def test_deploy_templates_list_pagination_no_limit(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.deploy_template.DeployTemplateManager(
            self.api)
        deploy_templates = self.mgr.list(limit=0)
        expect = [
            ('GET', '/v1/deploy_templates', {}, None),
            ('GET', '/v1/deploy_templates/?limit=1', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(deploy_templates, HasLength(2))

    def test_deploy_templates_list_sort_key(self) -> None:
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = ironicclient.v1.deploy_template.DeployTemplateManager(
            self.api)
        deploy_templates = self.mgr.list(sort_key='updated_at')
        expect = [
            ('GET', '/v1/deploy_templates/?sort_key=updated_at', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(deploy_templates))

    def test_deploy_templates_list_sort_dir(self) -> None:
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = ironicclient.v1.deploy_template.DeployTemplateManager(
            self.api)
        deploy_templates = self.mgr.list(sort_dir='desc')
        expect = [
            ('GET', '/v1/deploy_templates/?sort_dir=desc', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(deploy_templates))

    def test_deploy_templates_show(self) -> None:
        deploy_template = self.mgr.get(DEPLOY_TEMPLATE['uuid'])
        expect = [
            ('GET', '/v1/deploy_templates/%s' % DEPLOY_TEMPLATE['uuid'], {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(DEPLOY_TEMPLATE['uuid'], deploy_template.uuid)
        self.assertEqual(DEPLOY_TEMPLATE['name'], deploy_template.name)
        self.assertEqual(DEPLOY_TEMPLATE['steps'], deploy_template.steps)
        self.assertEqual(DEPLOY_TEMPLATE['extra'], deploy_template.extra)

    def test_deploy_template_show_fields(self) -> None:
        deploy_template = self.mgr.get(DEPLOY_TEMPLATE['uuid'],
                                       fields=['uuid', 'name'])
        expect = [
            ('GET', '/v1/deploy_templates/%s?fields=uuid,name' %
             DEPLOY_TEMPLATE['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(DEPLOY_TEMPLATE['uuid'], deploy_template.uuid)
        self.assertEqual(DEPLOY_TEMPLATE['name'], deploy_template.name)

    def test_create(self) -> None:
        deploy_template = self.mgr.create(**CREATE_DEPLOY_TEMPLATE)
        expect = [
            ('POST', '/v1/deploy_templates', {}, CREATE_DEPLOY_TEMPLATE),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(deploy_template)

    def test_create_with_uuid(self) -> None:
        deploy_template = self.mgr.create(**CREATE_DEPLOY_TEMPLATE_WITH_UUID)
        expect = [
            ('POST', '/v1/deploy_templates', {},
             CREATE_DEPLOY_TEMPLATE_WITH_UUID),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(deploy_template)

    def test_delete(self) -> None:
        self.mgr.delete(template_id=DEPLOY_TEMPLATE['uuid'])
        expect = [
            ('DELETE', '/v1/deploy_templates/%s' % DEPLOY_TEMPLATE['uuid'], {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)

    def test_update(self) -> None:
        patch = {'op': 'replace',
                 'value': NEW_NAME,
                 'path': '/name'}
        deploy_template = self.mgr.update(
            template_id=DEPLOY_TEMPLATE['uuid'], patch=patch)
        expect = [
            ('PATCH', '/v1/deploy_templates/%s' % DEPLOY_TEMPLATE['uuid'],
             {}, patch),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(NEW_NAME, deploy_template.name)



#  Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from __future__ import annotations

import copy

from osc_lib.tests import utils as oscutils

from ironicclient.osc.v1 import baremetal_driver
from ironicclient.tests.unit.osc.v1 import fakes as baremetal_fakes


class TestBaremetalDriver(baremetal_fakes.TestBaremetal):

    def setUp(self) -> None:
        super(TestBaremetalDriver, self).setUp()

        self.baremetal_mock = self.app.client_manager.baremetal
        self.baremetal_mock.reset_mock()


class TestListBaremetalDriver(TestBaremetalDriver):

    def setUp(self) -> None:
        super(TestListBaremetalDriver, self).setUp()

        self.baremetal_mock.driver.list.return_value = [
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.BAREMETAL_DRIVER),
                loaded=True)
        ]
        self.cmd = baremetal_driver.ListBaremetalDriver(self.app, None)

    def test_baremetal_driver_list(self) -> None:
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        collist = (
            "name",
            "hosts")
        self.assertEqual(collist, tuple(columns))

        datalist = ((
            baremetal_fakes.baremetal_driver_name,
            ', '.join(baremetal_fakes.baremetal_driver_hosts)), )
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_driver_list_with_type(self) -> None:
        arglist = ['--type', baremetal_fakes.baremetal_driver_type]
        verifylist = [('type', baremetal_fakes.baremetal_driver_type)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        collist = (
            "name",
            "hosts")
        self.assertEqual(collist, tuple(columns))

        datalist = ((
            baremetal_fakes.baremetal_driver_name,
            ', '.join(baremetal_fakes.baremetal_driver_hosts)),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_driver_list_with_detail(self) -> None:
        arglist = ['--long']
        verifylist = [('long', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        collist = (
            "name",
            "type",
            "hosts",
            'default_bios_interface',
            'default_boot_interface',
            'default_console_interface',
            'default_deploy_interface',
            'default_firmware_interface',
            'default_inspect_interface',
            'default_management_interface',
            'default_network_interface',
            'default_power_interface',
            'default_raid_interface',
            'default_rescue_interface',
            'default_storage_interface',
            'default_vendor_interface',
            'enabled_bios_interfaces',
            'enabled_boot_interfaces',
            'enabled_console_interfaces',
            'enabled_deploy_interfaces',
            'enabled_firmware_interfaces',
            'enabled_inspect_interfaces',
            'enabled_management_interfaces',
            'enabled_network_interfaces',
            'enabled_power_interfaces',
            'enabled_raid_interfaces',
            'enabled_rescue_interfaces',
            'enabled_storage_interfaces',
            'enabled_vendor_interfaces'
        )
        self.assertEqual(collist, tuple(columns))

        datalist = ((
            baremetal_fakes.baremetal_driver_name,
            baremetal_fakes.baremetal_driver_type,
            ', '.join(baremetal_fakes.baremetal_driver_hosts),
            baremetal_fakes.baremetal_driver_default_bios_if,
            baremetal_fakes.baremetal_driver_default_boot_if,
            baremetal_fakes.baremetal_driver_default_console_if,
            baremetal_fakes.baremetal_driver_default_deploy_if,
            baremetal_fakes.baremetal_driver_default_firmware_if,
            baremetal_fakes.baremetal_driver_default_inspect_if,
            baremetal_fakes.baremetal_driver_default_management_if,
            baremetal_fakes.baremetal_driver_default_network_if,
            baremetal_fakes.baremetal_driver_default_power_if,
            baremetal_fakes.baremetal_driver_default_raid_if,
            baremetal_fakes.baremetal_driver_default_rescue_if,
            baremetal_fakes.baremetal_driver_default_storage_if,
            baremetal_fakes.baremetal_driver_default_vendor_if,
            ', '.join(baremetal_fakes.baremetal_driver_enabled_bios_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_boot_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_console_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_deploy_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_firmware_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_inspect_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_management_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_network_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_power_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_raid_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_rescue_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_storage_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_vendor_ifs),
        ),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_driver_list_fields(self) -> None:
        arglist = [
            '--fields', 'name', 'hosts'
        ]
        verifylist = [
            ('fields', [['name', 'hosts']])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'driver_type': None,
            'detail': None,
            'fields': ('name', 'hosts')
        }

        self.baremetal_mock.driver.list.assert_called_with(**kwargs)

    def test_baremetal_driver_list_fields_multiple(self) -> None:
        arglist = [
            '--fields', 'name',
            '--fields', 'hosts', 'type'
        ]
        verifylist = [
            ('fields', [['name'], ['hosts', 'type']])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'driver_type': None,
            'detail': None,
            'fields': ('name', 'hosts', 'type')
        }

        self.baremetal_mock.driver.list.assert_called_with(**kwargs)

    def test_baremetal_driver_list_invalid_fields(self) -> None:
        arglist = [
            '--fields', 'name', 'invalid'
        ]
        verifylist = [
            ('fields', [['name', 'invalid']])
        ]
        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_driver_list_fields_with_long(self) -> None:
        arglist = [
            '--fields', 'name', 'hosts',
            '--long'
        ]
        verifylist = [
            ('fields', [['name', 'invalid']]),
            ('long', True)
        ]
        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestListBaremetalDriverProperty(TestBaremetalDriver):

    def setUp(self) -> None:
        super(TestListBaremetalDriverProperty, self).setUp()

        self.baremetal_mock.driver.properties.return_value = {
            'property1': 'description1', 'property2': 'description2'}
        self.cmd = baremetal_driver.ListBaremetalDriverProperty(
            self.app, None)

    def test_baremetal_driver_property_list(self) -> None:
        arglist = ['fakedrivername']
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.baremetal_mock.driver.properties.assert_called_with(*arglist)

        collist = ['property', 'description']
        self.assertEqual(collist, columns)
        expected_data = [('property1', 'description1'),
                         ('property2', 'description2')]
        self.assertEqual(expected_data, data)

    def test_baremetal_driver_list_no_arg(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestListBaremetalDriverRaidProperty(TestBaremetalDriver):

    def setUp(self) -> None:
        super(TestListBaremetalDriverRaidProperty, self).setUp()

        (self.baremetal_mock.driver.
         raid_logical_disk_properties.return_value) = {
            'RAIDProperty1': 'driver_raid_property1',
            'RAIDProperty2': 'driver_raid_property2',
        }

        self.cmd = (
            baremetal_driver.ListBaremetalDriverRaidProperty(
                self.app, None))

    def test_baremetal_driver_raid_property_list(self) -> None:
        arglist = ['fakedrivername']
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        (self.baremetal_mock.driver.
         raid_logical_disk_properties.assert_called_with(*arglist))

        collist = ('property', 'description')
        self.assertEqual(collist, tuple(columns))

        expected_data = [('RAIDProperty1', 'driver_raid_property1'),
                         ('RAIDProperty2', 'driver_raid_property2')]
        self.assertEqual(expected_data, data)

    def test_baremetal_driver_raid_property_list_no_arg(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestPassthruCallBaremetalDriver(TestBaremetalDriver):

    def setUp(self) -> None:
        super(TestPassthruCallBaremetalDriver, self).setUp()

        self.baremetal_mock.driver.vendor_passthru.return_value = (
            baremetal_fakes.BAREMETAL_DRIVER_PASSTHRU
        )

        self.cmd = baremetal_driver.PassthruCallBaremetalDriver(self.app, None)

    def test_baremetal_driver_passthru_call_with_min_args(self) -> None:

        arglist = [
            baremetal_fakes.baremetal_driver_name,
            baremetal_fakes.baremetal_driver_passthru_method,
        ]

        verifylist = [
            ('driver', baremetal_fakes.baremetal_driver_name),
            ('method', baremetal_fakes.baremetal_driver_passthru_method),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        # Set expected values
        args = [
            baremetal_fakes.baremetal_driver_name,
            baremetal_fakes.baremetal_driver_passthru_method,
        ]
        kwargs = {
            'http_method': 'POST',
            'args': {}
        }
        (self.baremetal_mock.driver.vendor_passthru.
            assert_called_once_with(*args, **kwargs))

    def test_baremetal_driver_passthru_call_with_all_args(self) -> None:

        arglist = [
            baremetal_fakes.baremetal_driver_name,
            baremetal_fakes.baremetal_driver_passthru_method,
            '--arg', 'arg1=val1', '--arg', 'arg2=val2',
            '--http-method', 'POST'
        ]

        verifylist = [
            ('driver', baremetal_fakes.baremetal_driver_name),
            ('method', baremetal_fakes.baremetal_driver_passthru_method),
            ('arg', ['arg1=val1', 'arg2=val2']),
            ('http_method', 'POST')
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        # Set expected values
        args = [
            baremetal_fakes.baremetal_driver_name,
            baremetal_fakes.baremetal_driver_passthru_method,
        ]
        kwargs = {
            'http_method': 'POST',
            'args': {'arg1': 'val1', 'arg2': 'val2'}
        }
        (self.baremetal_mock.driver.vendor_passthru.
            assert_called_once_with(*args, **kwargs))

    def test_baremetal_driver_passthru_call_no_arg(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestPassthruListBaremetalDriver(TestBaremetalDriver):

    def setUp(self) -> None:
        super(TestPassthruListBaremetalDriver, self).setUp()

        self.baremetal_mock.driver.get_vendor_passthru_methods.return_value = (
            baremetal_fakes.BAREMETAL_DRIVER_PASSTHRU
        )
        self.cmd = baremetal_driver.PassthruListBaremetalDriver(self.app, None)

    def test_baremetal_driver_passthru_list(self) -> None:
        arglist = ['fakedrivername']
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        args = ['fakedrivername']
        (self.baremetal_mock.driver.get_vendor_passthru_methods.
            assert_called_with(*args))

        collist = (
            "name",
            "http_methods",
            "async",
            "description",
            "attach",
        )
        self.assertEqual(collist, tuple(columns))

        datalist = (('lookup', 'POST', 'false', '', 'false'),)

        self.assertEqual(datalist, tuple(data))

    def test_baremetal_driver_passthru_list_no_arg(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestShowBaremetalDriver(TestBaremetalDriver):

    def setUp(self) -> None:
        super(TestShowBaremetalDriver, self).setUp()

        self.baremetal_mock.driver.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.BAREMETAL_DRIVER),
                loaded=True))
        self.cmd = baremetal_driver.ShowBaremetalDriver(self.app, None)

    def test_baremetal_driver_show(self) -> None:
        arglist = ['fakedrivername']
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        args = ['fakedrivername']
        self.baremetal_mock.driver.get.assert_called_with(*args, fields=None)
        self.assertFalse(self.baremetal_mock.driver.properties.called)

        collist = ('default_bios_interface', 'default_boot_interface',
                   'default_console_interface', 'default_deploy_interface',
                   'default_firmware_interface', 'default_inspect_interface',
                   'default_management_interface',
                   'default_network_interface', 'default_power_interface',
                   'default_raid_interface', 'default_rescue_interface',
                   'default_storage_interface', 'default_vendor_interface',
                   'enabled_bios_interfaces', 'enabled_boot_interfaces',
                   'enabled_console_interfaces', 'enabled_deploy_interfaces',
                   'enabled_firmware_interfaces',
                   'enabled_inspect_interfaces',
                   'enabled_management_interfaces',
                   'enabled_network_interfaces', 'enabled_power_interfaces',
                   'enabled_raid_interfaces', 'enabled_rescue_interfaces',
                   'enabled_storage_interfaces', 'enabled_vendor_interfaces',
                   'hosts', 'name', 'type')
        self.assertEqual(collist, columns)

        datalist = (
            baremetal_fakes.baremetal_driver_default_bios_if,
            baremetal_fakes.baremetal_driver_default_boot_if,
            baremetal_fakes.baremetal_driver_default_console_if,
            baremetal_fakes.baremetal_driver_default_deploy_if,
            baremetal_fakes.baremetal_driver_default_firmware_if,
            baremetal_fakes.baremetal_driver_default_inspect_if,
            baremetal_fakes.baremetal_driver_default_management_if,
            baremetal_fakes.baremetal_driver_default_network_if,
            baremetal_fakes.baremetal_driver_default_power_if,
            baremetal_fakes.baremetal_driver_default_raid_if,
            baremetal_fakes.baremetal_driver_default_rescue_if,
            baremetal_fakes.baremetal_driver_default_storage_if,
            baremetal_fakes.baremetal_driver_default_vendor_if,
            ', '.join(baremetal_fakes.baremetal_driver_enabled_bios_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_boot_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_console_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_deploy_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_firmware_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_inspect_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_management_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_network_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_power_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_raid_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_rescue_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_storage_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_enabled_vendor_ifs),
            ', '.join(baremetal_fakes.baremetal_driver_hosts),
            baremetal_fakes.baremetal_driver_name,
            baremetal_fakes.baremetal_driver_type)

        self.assertEqual(datalist, tuple(data))

    def test_baremetal_driver_show_no_arg(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_driver_show_fields(self) -> None:
        arglist = [
            'fakedrivername',
            '--fields', 'name', 'hosts'
        ]
        verifylist = [
            ('driver', 'fakedrivername'),
            ('fields', [['name', 'hosts']])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        args = ['fakedrivername']
        fields = ['name', 'hosts']
        self.baremetal_mock.driver.get.assert_called_with(*args, fields=fields)
        self.assertFalse(self.baremetal_mock.driver.properties.called)

    def test_baremetal_driver_show_fields_multiple(self) -> None:
        arglist = [
            'fakedrivername',
            '--fields', 'name',
            '--fields', 'hosts', 'type'
        ]
        verifylist = [
            ('driver', 'fakedrivername'),
            ('fields', [['name'], ['hosts', 'type']])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        args = ['fakedrivername']
        fields = ['name', 'hosts', 'type']
        self.baremetal_mock.driver.get.assert_called_with(*args, fields=fields)
        self.assertFalse(self.baremetal_mock.driver.properties.called)

    def test_baremetal_driver_show_invalid_fields(self) -> None:
        arglist = [
            'fakedrivername',
            '--fields', 'name', 'invalid'
        ]
        verifylist = [
            ('driver', 'fakedrivername'),
            ('fields', [['name', 'invalid']])
        ]

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)



#
# Copyright 2016 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from ironicclient.common import base


class Event(base.Resource):
    def __repr__(self) -> str:
        return "<Event: %s>" % self.name


class EventManager(base.CreateManager[Event]):
    resource_class: type[Event] = Event
    _creation_attributes: list[str] = ['events']
    _resource_name: str = 'events'



#
#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import json
from unittest import mock

from osc_lib.tests import utils

from ironicclient.tests.unit.osc import fakes

baremetal_chassis_uuid = 'aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee'
baremetal_chassis_uuid_empty = ''
baremetal_properties_empty = ''
baremetal_chassis_description = 'chassis description'
baremetal_chassis_extra = {}
BAREMETAL_CHASSIS = {
    'uuid': baremetal_chassis_uuid,
    'description': baremetal_chassis_description,
    'extra': baremetal_chassis_extra,
}

baremetal_uuid = 'xxx-xxxxxx-xxxx'
baremetal_name = 'fake name'
baremetal_owner = 'fake-owner'
baremetal_instance_uuid = 'yyy-yyyyyy-yyyy'
baremetal_power_state = None
baremetal_provision_state = None
baremetal_maintenance = False

BAREMETAL = {
    'uuid': baremetal_uuid,
    'name': baremetal_name,
    'instance_uuid': baremetal_instance_uuid,
    'power_state': baremetal_power_state,
    'provision_state': baremetal_provision_state,
    'maintenance': baremetal_maintenance,
    'links': [],
    'volume': [],
}

baremetal_port_uuid = 'zzz-zzzzzz-zzzz'
baremetal_port_address = 'AA:BB:CC:DD:EE:FF'
baremetal_port_extra = {'key1': 'value1',
                        'key2': 'value2'}
baremetal_port_physical_network = 'physnet1'

BAREMETAL_PORT = {
    'uuid': baremetal_port_uuid,
    'address': baremetal_port_address,
    'extra': baremetal_port_extra,
    'node_uuid': baremetal_uuid,
}

baremetal_driver_hosts = ['fake-host1', 'fake-host2']
baremetal_driver_name = 'fakedrivername'
baremetal_driver_type = 'classic'
baremetal_driver_default_bios_if = 'bios'
baremetal_driver_default_boot_if = 'boot'
baremetal_driver_default_console_if = 'console'
baremetal_driver_default_deploy_if = 'deploy'
baremetal_driver_default_inspect_if = 'inspect'
baremetal_driver_default_firmware_if = 'firmware'
baremetal_driver_default_management_if = 'management'
baremetal_driver_default_network_if = 'network'
baremetal_driver_default_power_if = 'power'
baremetal_driver_default_raid_if = 'raid'
baremetal_driver_default_rescue_if = 'rescue'
baremetal_driver_default_storage_if = 'storage'
baremetal_driver_default_vendor_if = 'vendor'
baremetal_driver_enabled_bios_ifs = ['bios', 'bios2']
baremetal_driver_enabled_boot_ifs = ['boot', 'boot2']
baremetal_driver_enabled_console_ifs = ['console', 'console2']
baremetal_driver_enabled_deploy_ifs = ['deploy', 'deploy2']
baremetal_driver_enabled_firmware_ifs = ['firmware', 'firmware2']
baremetal_driver_enabled_inspect_ifs = ['inspect', 'inspect2']
baremetal_driver_enabled_management_ifs = ['management', 'management2']
baremetal_driver_enabled_network_ifs = ['network', 'network2']
baremetal_driver_enabled_power_ifs = ['power', 'power2']
baremetal_driver_enabled_raid_ifs = ['raid', 'raid2']
baremetal_driver_enabled_rescue_ifs = ['rescue', 'rescue2']
baremetal_driver_enabled_storage_ifs = ['storage', 'storage2']
baremetal_driver_enabled_vendor_ifs = ['vendor', 'vendor2']

BAREMETAL_DRIVER = {
    'hosts': baremetal_driver_hosts,
    'name': baremetal_driver_name,
    'type': baremetal_driver_type,
    'default_bios_interface': baremetal_driver_default_bios_if,
    'default_boot_interface': baremetal_driver_default_boot_if,
    'default_console_interface': baremetal_driver_default_console_if,
    'default_deploy_interface': baremetal_driver_default_deploy_if,
    'default_firmware_interface': baremetal_driver_default_firmware_if,
    'default_inspect_interface': baremetal_driver_default_inspect_if,
    'default_management_interface': baremetal_driver_default_management_if,
    'default_network_interface': baremetal_driver_default_network_if,
    'default_power_interface': baremetal_driver_default_power_if,
    'default_raid_interface': baremetal_driver_default_raid_if,
    'default_rescue_interface': baremetal_driver_default_rescue_if,
    'default_storage_interface': baremetal_driver_default_storage_if,
    'default_vendor_interface': baremetal_driver_default_vendor_if,
    'enabled_bios_interfaces': baremetal_driver_enabled_bios_ifs,
    'enabled_boot_interfaces': baremetal_driver_enabled_boot_ifs,
    'enabled_console_interfaces': baremetal_driver_enabled_console_ifs,
    'enabled_deploy_interfaces': baremetal_driver_enabled_deploy_ifs,
    'enabled_firmware_interfaces': baremetal_driver_enabled_firmware_ifs,
    'enabled_inspect_interfaces': baremetal_driver_enabled_inspect_ifs,
    'enabled_management_interfaces': baremetal_driver_enabled_management_ifs,
    'enabled_network_interfaces': baremetal_driver_enabled_network_ifs,
    'enabled_power_interfaces': baremetal_driver_enabled_power_ifs,
    'enabled_raid_interfaces': baremetal_driver_enabled_raid_ifs,
    'enabled_rescue_interfaces': baremetal_driver_enabled_rescue_ifs,
    'enabled_storage_interfaces': baremetal_driver_enabled_storage_ifs,
    'enabled_vendor_interfaces': baremetal_driver_enabled_vendor_ifs,
}

baremetal_driver_passthru_method = 'lookup'

BAREMETAL_DRIVER_PASSTHRU = {"lookup": {"attach": "false",
                                        "http_methods": ["POST"],
                                        "description": "",
                                        "async": "false"}}

baremetal_portgroup_uuid = 'ppp-gggggg-pppp'
baremetal_portgroup_name = 'Portgroup-name'
baremetal_portgroup_address = 'AA:BB:CC:CC:BB:AA'
baremetal_portgroup_mode = 'active-backup'
baremetal_portgroup_extra = {'key1': 'value1',
                             'key2': 'value2'}
baremetal_portgroup_properties = {'key1': 'value11',
                                  'key2': 'value22'}
baremetal_portgroup_category = 'Green'

PORTGROUP = {'uuid': baremetal_portgroup_uuid,
             'name': baremetal_portgroup_name,
             'node_uuid': baremetal_uuid,
             'address': baremetal_portgroup_address,
             'extra': baremetal_portgroup_extra,
             'mode': baremetal_portgroup_mode,
             'properties': baremetal_portgroup_properties,
             'physical_network': baremetal_port_physical_network,
             'category': baremetal_portgroup_category,
             }

VIFS = {'vifs': [{'id': 'aaa-aa'}]}
TRAITS = ['CUSTOM_FOO', 'CUSTOM_BAR']
CHILDREN = ['53da080f-6de7-4a3e-bcb6-b7889b380ad0',
            '48467e9b-3cd1-45b5-a57e-169e01370169']
BIOS_SETTINGS = [{'name': 'bios_name_1', 'value': 'bios_value_1', 'links': []},
                 {'name': 'bios_name_2', 'value': 'bios_value_2', 'links': []}]

BIOS_DETAILED_SETTINGS = [{'name': 'SysName', 'value': 'my-system',
                           'links': [], 'attribute_type': 'String',
                           'min_length': '1', 'max_length': '16'},
                          {'name': 'NumCores', 'value': '10',
                           'links': [], 'attribute_type': 'Integer',
                           'lower_bound': '10', 'upper_bound': '20'},
                          {'name': 'ProcVirtualization', 'value': 'Enabled',
                           'links': [], 'attribute_type': 'Enumeration',
                           'allowable_values': ['Enabled', 'Disabled']}]

baremetal_volume_connector_uuid = 'vvv-cccccc-vvvv'
baremetal_volume_connector_type = 'iqn'
baremetal_volume_connector_connector_id = 'iqn.2017-01.connector'
baremetal_volume_connector_extra = {'key1': 'value1',
                                    'key2': 'value2'}
VOLUME_CONNECTOR = {
    'uuid': baremetal_volume_connector_uuid,
    'node_uuid': baremetal_uuid,
    'type': baremetal_volume_connector_type,
    'connector_id': baremetal_volume_connector_connector_id,
    'extra': baremetal_volume_connector_extra,
}

baremetal_volume_target_uuid = 'vvv-tttttt-vvvv'
baremetal_volume_target_volume_type = 'iscsi'
baremetal_volume_target_boot_index = 0
baremetal_volume_target_volume_id = 'vvv-tttttt-iii'
baremetal_volume_target_extra = {'key1': 'value1',
                                 'key2': 'value2'}
baremetal_volume_target_properties = {'key11': 'value11',
                                      'key22': 'value22'}
VOLUME_TARGET = {
    'uuid': baremetal_volume_target_uuid,
    'node_uuid': baremetal_uuid,
    'volume_type': baremetal_volume_target_volume_type,
    'boot_index': baremetal_volume_target_boot_index,
    'volume_id': baremetal_volume_target_volume_id,
    'extra': baremetal_volume_target_extra,
    'properties': baremetal_volume_target_properties,
}

baremetal_hostname = 'compute1.localdomain'
baremetal_conductor_group = 'foo'
baremetal_alive = True
baremetal_drivers = ['fake-hardware']
CONDUCTOR = {
    'hostname': baremetal_hostname,
    'conductor_group': baremetal_conductor_group,
    'alive': baremetal_alive,
    'drivers': baremetal_drivers,
}

baremetal_allocation_state = 'active'
baremetal_resource_class = 'baremetal'
ALLOCATION = {
    'resource_class': baremetal_resource_class,
    'uuid': baremetal_uuid,
    'name': baremetal_name,
    'state': baremetal_allocation_state,
    'node_uuid': baremetal_uuid,
}

baremetal_deploy_template_uuid = 'ddd-tttttt-dddd'
baremetal_deploy_template_name = 'DeployTemplate-name'
baremetal_deploy_template_steps = json.dumps([{
    'interface': 'raid',
    'step': 'create_configuration',
    'args': {},
    'priority': 10
}])
baremetal_deploy_template_extra = {'key1': 'value1', 'key2': 'value2'}
DEPLOY_TEMPLATE = {
    'uuid': baremetal_deploy_template_uuid,
    'name': baremetal_deploy_template_name,
    'steps': baremetal_deploy_template_steps,
    'extra': baremetal_deploy_template_extra,
}

baremetal_runbook_uuid = 'ddd-tttttt-dddd'
baremetal_runbook_name = 'CUSTOM_AWESOME'
baremetal_runbook_owner = 'some_user'
baremetal_runbook_public = False
baremetal_runbook_steps = json.dumps([{
    'interface': 'raid',
    'step': 'create_configuration',
    'args': {},
    'order': 1
}])
baremetal_runbook_extra = {'key1': 'value1', 'key2': 'value2'}
baremetal_runbook_description = ''
baremetal_runbook_traits = []
RUNBOOK = {
    'uuid': baremetal_runbook_uuid,
    'name': baremetal_runbook_name,
    'description': baremetal_runbook_description,
    'traits': baremetal_runbook_traits,
    'owner': baremetal_runbook_owner,
    'public': baremetal_runbook_public,
    'steps': baremetal_runbook_steps,
    'extra': baremetal_runbook_extra,
}

baremetal_shard_name = 'example_shard'
baremetal_shard_count = 47
SHARD = {
    'name': baremetal_shard_name,
    'count': baremetal_shard_count,
}

NODE_HISTORY = [
    {
        'uuid': 'abcdef1',
        'created_at': 'time',
        'severity': 'info',
        'event': 'meow',
        'event_type': 'purring',
        'conductor': 'lap-conductor',
        'user': '0191',
        'links': {'href': 'url', 'rel': 'self'},
    }
]
NODE_INVENTORY = [
    {
        'inventory':
        {
            'memory': {'physical_mb': 3072},
            'cpu': {'count': 1,
                    'model_name': 'qemu64',
                    'architecture': 'x86_64'},
            'disks': [{'name': 'testvm2.qcow2',
                       'size': 11811160064}],
            'interfaces': [{'mac_address': '52:54:00:11:2d:26'}],
            'system_vendor': {'product_name': 'testvm2',
                              'manufacturer': 'Sushy Emulator'},
            'boot': {'current_boot_mode': 'uefi'}
        }
    }
]

FIRMWARE_COMPONENTS = [
    {
        "component": "bios",
        "initial_version": "v1.0.0.0 (01.02.2022)",
        "current_version": "v1.2.3.4 (01.02.2023)",
        "last_version_flashed": "v1.2.3.4 (01.02.2023)",
        "created_at": "2023-02-01 09:00:00",
        "updated_at": "2023-03-01 10:00:00"
    },
    {
        "component": "bmc",
        "initial_version": "v1.0.0",
        "current_version": "v1.0.0",
        "last_version_flashed": "",
        "created_at": "2023-02-01 09:00:00",
        "updated_at": ""
    }
]


baremetal_inspection_rule_uuid = 'ddd-tttttt-dddd'
baremetal_inspection_rule_description = 'Blah'
baremetal_inspection_rule_priority = 0
baremetal_inspection_rule_sensitive = False
baremetal_inspection_rule_phase = 'main'
baremetal_inspection_rule_actions = json.dumps([{
    'op': 'set-attribute',
    'args': ["/driver", "idrac"],
}])
baremetal_inspection_rule_conditions = json.dumps([{
    'op': 'is-true',
    'args': ["{node.auto_discovered}"],
    'multiple': 'any',
}])

INSPECTION_RULE = {
    'uuid': baremetal_inspection_rule_uuid,
    'description': baremetal_inspection_rule_description,
    'priority': baremetal_inspection_rule_priority,
    'sensitive': baremetal_inspection_rule_sensitive,
    'phase': baremetal_inspection_rule_phase,
    'actions': baremetal_inspection_rule_actions,
    'conditions': baremetal_inspection_rule_conditions,
}


class TestBaremetal(utils.TestCommand):

    def setUp(self) -> None:
        super(TestBaremetal, self).setUp()

        # TODO(anandkaranubc): drop setattr once osc_lib annotates
        # FakeClientManager.auth_ref as ``object | None`` (currently inferred
        # as ``None``, so direct assignment trips ``[assignment]``).
        setattr(self.app.client_manager, 'auth_ref',
                mock.Mock(auth_token="TOKEN"))
        self.app.client_manager.baremetal = mock.Mock()


class FakeBaremetalResource(fakes.FakeResource):

    def get_keys(self) -> dict[str, str]:
        return {'property': 'value'}



#
#   Copyright 2016 Intel Corporation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import copy
from unittest import mock

from osc_lib.tests import utils as oscutils

from ironicclient import exc
from ironicclient.osc.v1 import baremetal_chassis
from ironicclient.tests.unit.osc.v1 import fakes as baremetal_fakes


class TestChassis(baremetal_fakes.TestBaremetal):

    def setUp(self) -> None:
        super(TestChassis, self).setUp()

        # Get a shortcut to the baremetal manager mock
        self.baremetal_mock = self.app.client_manager.baremetal
        self.baremetal_mock.reset_mock()


class TestChassisCreate(TestChassis):
    def setUp(self) -> None:
        super(TestChassisCreate, self).setUp()

        self.baremetal_mock.chassis.create.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.BAREMETAL_CHASSIS),
                loaded=True,
            ))

        # Get the command object to test
        self.cmd = baremetal_chassis.CreateBaremetalChassis(self.app, None)
        self.arglist = []
        self.verifylist = []
        self.collist = (
            'uuid',
            'description',
            'extra',
        )
        self.datalist = (
            baremetal_fakes.baremetal_chassis_uuid,
            baremetal_fakes.baremetal_chassis_description,
            baremetal_fakes.baremetal_chassis_extra,
        )
        self.actual_kwargs = {}

    def check_with_options(
            self,
            addl_arglist: list[str],
            addl_verifylist: list[tuple[str, object]],
            addl_kwargs: dict[str, object],
    ) -> None:
        arglist = copy.copy(self.arglist) + addl_arglist
        verifylist = copy.copy(self.verifylist) + addl_verifylist

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        collist = copy.copy(self.collist)
        self.assertEqual(collist, columns)
        self.assertNotIn('nodes', columns)

        datalist = copy.copy(self.datalist)
        self.assertEqual(datalist, tuple(data))

        kwargs = copy.copy(self.actual_kwargs)
        kwargs.update(addl_kwargs)

        self.baremetal_mock.chassis.create.assert_called_once_with(**kwargs)

    def test_chassis_create_no_options(self) -> None:
        self.check_with_options([], [], {})

    def test_chassis_create_with_description(self) -> None:
        description = 'chassis description'
        self.check_with_options(['--description', description],
                                [('description', description)],
                                {'description': description})

    def test_chassis_create_with_extra(self) -> None:
        extra1 = 'arg1=val1'
        extra2 = 'arg2=val2'
        self.check_with_options(['--extra', extra1,
                                 '--extra', extra2],
                                [('extra', [extra1, extra2])],
                                {'extra': {
                                    'arg1': 'val1',
                                    'arg2': 'val2'}})

    def test_chassis_create_with_uuid(self) -> None:
        uuid = baremetal_fakes.baremetal_chassis_uuid
        self.check_with_options(['--uuid', uuid],
                                [('uuid', uuid)],
                                {'uuid': uuid})


class TestChassisDelete(TestChassis):
    def setUp(self) -> None:
        super(TestChassisDelete, self).setUp()

        self.baremetal_mock.chassis.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.BAREMETAL_CHASSIS),
                loaded=True,
            ))

        # Get the command object to test
        self.cmd = baremetal_chassis.DeleteBaremetalChassis(self.app, None)

    def test_chassis_delete(self) -> None:
        uuid = baremetal_fakes.baremetal_chassis_uuid
        arglist = [uuid]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.baremetal_mock.chassis.delete.assert_called_with(uuid)

    def test_chassis_delete_multiple(self) -> None:
        uuid1 = 'aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee'
        uuid2 = '11111111-2222-3333-4444-555555555555'
        arglist = [uuid1, uuid2]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # Set expected values
        args = [uuid1, uuid2]
        self.baremetal_mock.chassis.delete.assert_has_calls(
            [mock.call(x) for x in args]
        )
        self.assertEqual(2, self.baremetal_mock.chassis.delete.call_count)

    def test_chassis_delete_multiple_with_failure(self) -> None:
        uuid1 = 'aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee'
        uuid2 = '11111111-2222-3333-4444-555555555555'
        arglist = [uuid1, uuid2]
        verifylist = []

        self.baremetal_mock.chassis.delete.side_effect = [
            '', exc.ClientException]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(exc.ClientException,
                          self.cmd.take_action,
                          parsed_args)

        # Set expected values
        args = [uuid1, uuid2]
        self.baremetal_mock.chassis.delete.assert_has_calls(
            [mock.call(x) for x in args]
        )
        self.assertEqual(2, self.baremetal_mock.chassis.delete.call_count)


class TestChassisList(TestChassis):

    def setUp(self) -> None:
        super(TestChassisList, self).setUp()

        self.baremetal_mock.chassis.list.return_value = [
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.BAREMETAL_CHASSIS),
                loaded=True,
            ),
        ]

        # Get the command object to test
        self.cmd = baremetal_chassis.ListBaremetalChassis(self.app, None)

    def test_chassis_list_no_options(self) -> None:
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'marker': None,
            'limit': None,
        }

        self.baremetal_mock.chassis.list.assert_called_with(
            **kwargs
        )

        collist = (
            "uuid",
            "description",
        )
        self.assertEqual(collist, columns)
        datalist = ((
            baremetal_fakes.baremetal_chassis_uuid,
            baremetal_fakes.baremetal_chassis_description,
        ), )
        self.assertEqual(datalist, tuple(data))

    def test_chassis_list_long(self) -> None:
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'detail': True,
            'marker': None,
            'limit': None,
        }

        self.baremetal_mock.chassis.list.assert_called_with(
            **kwargs
        )

        collist = ('uuid', 'description', 'created_at', 'updated_at', 'extra')
        self.assertEqual(collist, columns)
        datalist = ((
            baremetal_fakes.baremetal_chassis_uuid,
            baremetal_fakes.baremetal_chassis_description,
            '',
            '',
            baremetal_fakes.baremetal_chassis_extra,
        ), )
        self.assertEqual(datalist, tuple(data))

    def test_chassis_list_fields(self) -> None:
        arglist = [
            '--fields', 'uuid', 'extra',
        ]
        verifylist = [
            ('fields', [['uuid', 'extra']]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'marker': None,
            'limit': None,
            'detail': False,
            'fields': ('uuid', 'extra'),
        }

        self.baremetal_mock.chassis.list.assert_called_with(
            **kwargs
        )

    def test_chassis_list_fields_multiple(self) -> None:
        arglist = [
            '--fields', 'uuid', 'description',
            '--fields', 'extra',
        ]
        verifylist = [
            ('fields', [['uuid', 'description'], ['extra']])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'marker': None,
            'limit': None,
            'detail': False,
            'fields': ('uuid', 'description', 'extra')
        }

        self.baremetal_mock.chassis.list.assert_called_with(
            **kwargs
        )

    def test_chassis_list_invalid_fields(self) -> None:
        arglist = [
            '--fields', 'uuid', 'invalid'
        ]
        verifylist = [
            ('fields', [['uuid', 'invalid']])
        ]

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_chassis_list_long_and_fields(self) -> None:
        arglist = [
            '--long',
            '--fields', 'uuid', 'invalid'
        ]
        verifylist = [
            ('long', True),
            ('fields', [['uuid', 'invalid']])
        ]

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestChassisSet(TestChassis):
    def setUp(self) -> None:
        super(TestChassisSet, self).setUp()

        self.baremetal_mock.chassis.update.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.BAREMETAL_CHASSIS),
                loaded=True,
            ))

        # Get the command object to test
        self.cmd = baremetal_chassis.SetBaremetalChassis(self.app, None)

    def test_chassis_set_no_options(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_chassis_set_no_property(self) -> None:
        uuid = baremetal_fakes.baremetal_chassis_uuid
        arglist = [uuid]
        verifylist = [('chassis', uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.assertFalse(self.baremetal_mock.chassis.update.called)

    def test_chassis_set_description(self) -> None:
        description = 'new description'
        uuid = baremetal_fakes.baremetal_chassis_uuid
        arglist = [
            uuid,
            '--description', 'new description',
        ]
        verifylist = [
            ('chassis', uuid),
            ('description', description)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.baremetal_mock.chassis.update.assert_called_once_with(
            uuid,
            [{'path': '/description', 'value': description, 'op': 'add'}]
        )

    def test_chassis_set_extra(self) -> None:
        uuid = baremetal_fakes.baremetal_chassis_uuid
        extra = 'foo=bar'
        arglist = [
            uuid,
            '--extra', extra,
        ]
        verifylist = [
            ('chassis', uuid),
            ('extra', [extra])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.baremetal_mock.chassis.update.assert_called_once_with(
            uuid,
            [{'path': '/extra/foo', 'value': 'bar', 'op': 'add'}]
        )


class TestChassisShow(TestChassis):
    def setUp(self) -> None:
        super(TestChassisShow, self).setUp()

        self.baremetal_mock.chassis.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.BAREMETAL_CHASSIS),
                loaded=True,
            ))

        # Get the command object to test
        self.cmd = baremetal_chassis.ShowBaremetalChassis(self.app, None)

    def test_chassis_show(self) -> None:
        arglist = [baremetal_fakes.baremetal_chassis_uuid]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        args = [baremetal_fakes.baremetal_chassis_uuid]

        self.baremetal_mock.chassis.get.assert_called_with(
            *args, fields=None
        )

        collist = (
            'description',
            'extra',
            'uuid'
        )
        self.assertEqual(collist, columns)
        self.assertNotIn('nodes', columns)
        datalist = (
            baremetal_fakes.baremetal_chassis_description,
            baremetal_fakes.baremetal_chassis_extra,
            baremetal_fakes.baremetal_chassis_uuid,
        )
        self.assertEqual(datalist, tuple(data))

    def test_chassis_show_no_chassis(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_chassis_show_fields(self) -> None:
        uuid = baremetal_fakes.baremetal_chassis_uuid
        arglist = [
            uuid,
            '--fields', 'uuid', 'description',
        ]
        verifylist = [
            ('chassis', uuid),
            ('fields', [['uuid', 'description']]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # Set expected values
        args = [uuid]
        fields = ['uuid', 'description']

        self.baremetal_mock.chassis.get.assert_called_with(
            *args, fields=fields
        )

    def test_chassis_show_fields_multiple(self) -> None:
        uuid = baremetal_fakes.baremetal_chassis_uuid
        arglist = [
            uuid,
            '--fields', 'uuid', 'description',
            '--fields', 'extra',
        ]
        verifylist = [
            ('chassis', uuid),
            ('fields', [['uuid', 'description'], ['extra']])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # Set expected values
        args = [uuid]
        fields = ['uuid', 'description', 'extra']

        self.baremetal_mock.chassis.get.assert_called_with(
            *args, fields=fields
        )

    def test_chassis_show_invalid_fields(self) -> None:
        uuid = baremetal_fakes.baremetal_chassis_uuid
        arglist = [
            uuid,
            '--fields', 'uuid', 'invalid'
        ]
        verifylist = [
            ('chassis', uuid),
            ('fields', [['uuid', 'invalid']])
        ]

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestChassisUnset(TestChassis):
    def setUp(self) -> None:
        super(TestChassisUnset, self).setUp()

        self.baremetal_mock.chassis.update.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.BAREMETAL_CHASSIS),
                loaded=True,
            ))

        # Get the command object to test
        self.cmd = baremetal_chassis.UnsetBaremetalChassis(self.app, None)

    def test_chassis_unset_no_options(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_chassis_unset_no_property(self) -> None:
        uuid = baremetal_fakes.baremetal_chassis_uuid
        arglist = [uuid]
        verifylist = [('chassis', uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.assertFalse(self.baremetal_mock.chassis.update.called)

    def test_chassis_unset_description(self) -> None:
        uuid = baremetal_fakes.baremetal_chassis_uuid
        arglist = [
            uuid,
            '--description',
        ]
        verifylist = [
            ('chassis', uuid),
            ('description', True)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.baremetal_mock.chassis.update.assert_called_once_with(
            uuid,
            [{'path': '/description', 'op': 'remove'}]
        )

    def test_chassis_unset_extra(self) -> None:
        uuid = baremetal_fakes.baremetal_chassis_uuid
        arglist = [
            uuid,
            '--extra', 'foo',
        ]
        verifylist = [
            ('chassis', uuid),
            ('extra', ['foo'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.baremetal_mock.chassis.update.assert_called_once_with(
            uuid,
            [{'path': '/extra/foo', 'op': 'remove'}]
        )



#
#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import sys


AUTH_TOKEN = "foobar"
AUTH_URL = "http://0.0.0.0"
API_VERSION = '1.6'


class FakeApp(object):
    def __init__(self) -> None:
        _stdout = None
        self.client_manager = None
        self.stdin = sys.stdin
        self.stdout = _stdout or sys.stdout
        self.stderr = sys.stderr
        self.restapi = None


class FakeClientManager(object):
    def __init__(self) -> None:
        self.identity = None
        self.auth_ref = None
        self.interface = 'public'
        self._region_name = 'RegionOne'
        self.session = 'fake session'
        self._api_version = {'baremetal': API_VERSION}


class FakeResource(object):
    def __init__(
            self,
            manager: object | None,
            info: dict[str, object],
            loaded: bool = False,
    ) -> None:
        self.__name__ = type(self).__name__
        self.manager = manager
        self._info = info
        self._add_details(info)
        self._loaded = loaded

    def _add_details(self, info: dict[str, object]) -> None:
        for (k, v) in info.items():
            setattr(self, k, v)

    def __repr__(self) -> str:
        reprkeys = sorted(k for k in self.__dict__.keys()
                          if k[0] != '_' and k != 'manager')
        info = ", ".join("%s=%s" % (k, getattr(self, k)) for k in reprkeys)
        return "<%s %s>" % (self.__class__.__name__, info)



# Copyright 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from collections.abc import Iterable

from ironicclient.common.i18n import _


class Resource(object):
    """Resource class

    This class is used to manage the various fields that a resource (e.g.
    Chassis, Node, Port) contains.  An individual field consists of a
    'field_id' (key) and a 'label' (value).  The caller only provides the
    'field_ids' when instantiating the object.

    Ordering of the 'field_ids' will be preserved as specified by the caller.

    It also provides the ability to exclude some of these fields when they are
    being used for sorting.
    """

    FIELDS: dict[str, str] = {
        'address': 'Address',
        'alive': 'Alive',
        'allocation_uuid': 'Allocation UUID',
        'allowable_values': 'Allowable Values',
        'async': 'Async',
        'attribute_type': 'Attribute Type',
        'automated_clean': 'Automated Clean',
        'attach': 'Response is attachment',
        'boot_index': 'Boot Index',
        'boot_mode': 'Boot Mode',
        'candidate_nodes': 'Candidate Nodes',
        'chassis_uuid': 'Chassis UUID',
        'clean_step': 'Clean Step',
        'conductor': 'Conductor',
        'conductor_group': 'Conductor Group',
        'console_enabled': 'Console Enabled',
        'count': 'Count',
        'created_at': 'Created At',
        'default_bios_interface': 'Default BIOS Interface',
        'default_boot_interface': 'Default Boot Interface',
        'default_console_interface': 'Default Console Interface',
        'default_deploy_interface': 'Default Deploy Interface',
        'default_firmware_interface': 'Default Firmware Interface',
        'default_inspect_interface': 'Default Inspect Interface',
        'default_management_interface': 'Default Management Interface',
        'default_network_interface': 'Default Network Interface',
        'default_power_interface': 'Default Power Interface',
        'default_raid_interface': 'Default RAID Interface',
        'default_rescue_interface': 'Default Rescue Interface',
        'default_storage_interface': 'Default Storage Interface',
        'default_vendor_interface': 'Default Vendor Interface',
        'deploy_step': 'Deploy Step',
        'description': 'Description',
        'driver': 'Driver',
        'driver_info': 'Driver Info',
        'driver_internal_info': 'Driver Internal Info',
        'drivers': 'Drivers',
        'enabled_bios_interfaces': 'Enabled BIOS Interfaces',
        'enabled_boot_interfaces': 'Enabled Boot Interfaces',
        'enabled_console_interfaces': 'Enabled Console Interfaces',
        'enabled_deploy_interfaces': 'Enabled Deploy Interfaces',
        'enabled_firmware_interfaces': 'Enabled Firmware Interfaces',
        'enabled_inspect_interfaces': 'Enabled Inspect Interfaces',
        'enabled_management_interfaces': 'Enabled Management Interfaces',
        'enabled_network_interfaces': 'Enabled Network Interfaces',
        'enabled_power_interfaces': 'Enabled Power Interfaces',
        'enabled_raid_interfaces': 'Enabled RAID Interfaces',
        'enabled_rescue_interfaces': 'Enabled Rescue Interfaces',
        'enabled_storage_interfaces': 'Enabled Storage Interfaces',
        'enabled_vendor_interfaces': 'Enabled Vendor Interfaces',
        'extra': 'Extra',
        'event': 'Description of the event',
        'event_type': 'Event Origin Type',
        'hostname': 'Hostname',
        'hosts': 'Active host(s)',
        'http_methods': 'Supported HTTP methods',
        'inspection_finished_at': 'Inspection Finished At',
        'inspection_started_at': 'Inspection Started At',
        'instance_info': 'Instance Info',
        'instance_name': 'Instance Name',
        'instance_uuid': 'Instance UUID',
        'internal_info': 'Internal Info',
        'last_error': 'Last Error',
        'lessee': 'Lessee',
        'lower_bound': 'Lower Bound',
        'maintenance': 'Maintenance',
        'maintenance_reason': 'Maintenance Reason',
        'max_length': 'Maximum Length',
        'fault': 'Fault',
        'min_length': 'Minimum Length',
        'mode': 'Mode',
        'name': 'Name',
        'network_data': 'Network Configuration',
        'node_uuid': 'Node UUID',
        'owner': 'Owner',
        'power_state': 'Power State',
        'properties': 'Properties',
        'protected': 'Protected',
        'protected_reason': 'Protected Reason',
        'provision_state': 'Provisioning State',
        'provision_updated_at': 'Provision Updated At',
        'raid_config': 'Current RAID configuration',
        'read_only': 'Read Only',
        'reset_required': 'Reset Required',
        'reservation': 'Reservation',
        'resource_class': 'Resource Class',
        'retired': 'Retired',
        'retired_reason': 'Retired Reason',
        'secure_boot': 'Secure Boot',
        'state': 'State',
        'steps': 'Steps',
        'target_power_state': 'Target Power State',
        'target_provision_state': 'Target Provision State',
        'target_raid_config': 'Target RAID configuration',
        'traits': 'Traits',
        'type': 'Type',
        'updated_at': 'Updated At',
        'uuid': 'UUID',
        'value': 'Value',
        'volume_id': 'Volume ID',
        'volume_type': 'Driver Volume Type',
        'category': 'Category',
        'vendor': 'Vendor',
        'local_link_connection': 'Local Link Connection',
        'pxe_enabled': 'PXE boot enabled',
        'portgroup_uuid': 'Portgroup UUID',
        'bios_interface': 'BIOS Interface',
        'boot_interface': 'Boot Interface',
        'console_interface': 'Console Interface',
        'deploy_interface': 'Deploy Interface',
        'inspect_interface': 'Inspect Interface',
        'management_interface': 'Management Interface',
        'network_interface': 'Network Interface',
        'power_interface': 'Power Interface',
        'raid_interface': 'RAID Interface',
        'rescue_interface': 'Rescue Interface',
        'storage_interface': 'Storage Interface',
        'severity': 'Severity',
        'unique': 'Unique',
        'upper_bound': 'Upper Bound',
        'user': 'User',
        'vendor_interface': 'Vendor Interface',
        'standalone_ports_supported': 'Standalone Ports Supported',
        'physical_network': 'Physical Network',
        'id': 'ID',
        'connector_id': 'Connector ID',
        'is_smartnic': 'Is Smart NIC port',
        'parent_node': 'Parent Node',
        'children': 'Child Nodes',
        'firmware_interface': 'Firmware Interface',
        'health': 'Health',
        'public': 'Public',
        'disable_power_off': 'Disable Power Off',
        'priority': 'Priority',
        'sensitive': 'Sensitive',
        'phase': 'Phase',
        'actions': 'Actions',
        'conditions': 'Conditions'
    }

    def __init__(
        self,
        field_ids: list[str],
        sort_excluded: list[str] | None = None,
        override_labels: dict[str, str] | None = None,
    ) -> None:
        """Create a Resource object

        :param field_ids:  A list of strings that the Resource object will
                           contain.  Each string must match an existing key in
                           FIELDS.
        :param sort_excluded: Optional. A list of strings that will not be used
                              for sorting.  Must be a subset of 'field_ids'.
        :param override_labels: Optional. A dictionary, where key is a field ID
                                and value is the label to be used. If
                                unspecified, uses the labels associated with
                                the fields from global FIELDS.

        :raises: ValueError if sort_excluded or override_labels contains values
                 not in field_ids
        """
        def check_param_fields(
            param_name: str,
            param_fields: Iterable[str],
        ) -> None:
            not_existing = set(param_fields) - set(field_ids)
            if not_existing:
                raise ValueError(
                    _("%(param)s specified with value not contained in "
                      "field_ids.  Unknown value(s): %(unknown)s")
                    % {'param': param_name,
                       'unknown': ', '.join(not_existing)})

        if override_labels is None:
            override_labels = {}
        else:
            check_param_fields('override_labels', override_labels.keys())

        self._fields: tuple[str, ...] = tuple(field_ids)
        self._labels: tuple[str, ...] = tuple(
            [override_labels.get(x) or self.FIELDS[x]
             for x in field_ids])

        if sort_excluded is None:
            sort_excluded = []
        else:
            check_param_fields('sort_excluded', sort_excluded)
        self._sort_fields: tuple[str, ...] = tuple(
            [x for x in field_ids if x not in sort_excluded])
        self._sort_labels: tuple[str, ...] = tuple(
            [override_labels.get(x) or self.FIELDS[x]
             for x in self._sort_fields])

    @property
    def fields(self) -> tuple[str, ...]:
        return self._fields

    @property
    def labels(self) -> tuple[str, ...]:
        return self._labels

    @property
    def sort_fields(self) -> tuple[str, ...]:
        return self._sort_fields

    @property
    def sort_labels(self) -> tuple[str, ...]:
        return self._sort_labels


# Chassis
CHASSIS_DETAILED_RESOURCE = Resource(
    ['uuid',
     'description',
     'created_at',
     'updated_at',
     'extra',
     ],
    sort_excluded=['extra'])
CHASSIS_RESOURCE = Resource(
    ['uuid',
     'description',
     ])

# Nodes
# NOTE(dtantsur): the sorting of the list must follow the sorting for the
# corresponding headings, so some items (like raid_config) may seem out of
# order here.
NODE_DETAILED_RESOURCE = Resource(
    ['allocation_uuid',
     'automated_clean',
     'bios_interface',
     'boot_interface',
     'boot_mode',
     'chassis_uuid',
     'clean_step',
     'conductor',
     'conductor_group',
     'console_enabled',
     'console_interface',
     'created_at',
     'raid_config',
     'deploy_interface',
     'deploy_step',
     'description',
     'disable_power_off',
     'driver',
     'driver_info',
     'driver_internal_info',
     'extra',
     'fault',
     'firmware_interface',
     'health',
     'inspect_interface',
     'inspection_finished_at',
     'inspection_started_at',
     'instance_info',
     'instance_name',
     'instance_uuid',
     'last_error',
     'lessee',
     'maintenance',
     'maintenance_reason',
     'management_interface',
     'name',
     'network_data',
     'network_interface',
     'owner',
     'parent_node',
     'power_interface',
     'power_state',
     'properties',
     'protected',
     'protected_reason',
     'provision_updated_at',
     'provision_state',
     'raid_interface',
     'rescue_interface',
     'reservation',
     'resource_class',
     'retired',
     'retired_reason',
     'secure_boot',
     'storage_interface',
     'target_power_state',
     'target_provision_state',
     'target_raid_config',
     'traits',
     'uuid',
     'updated_at',
     'vendor_interface',
     ],
    sort_excluded=[
        'allocation_uuid',
        # The server cannot sort on "chassis_uuid" because it isn't a column in
        # the "nodes" database table. "chassis_id" is stored, but it is
        # internal to ironic. See bug #1443003 for more details.
        'chassis_uuid',
        'clean_step',
        'deploy_step',
        'driver_info',
        'driver_internal_info',
        'extra',
        'instance_info',
        'properties',
        'raid_config',
        'target_raid_config',
        'traits',
    ])
NODE_RESOURCE = Resource(
    ['uuid',
     'name',
     'instance_uuid',
     'power_state',
     'provision_state',
     'maintenance',
     ])
VENDOR_PASSTHRU_METHOD_RESOURCE = Resource(
    ['name',
     'http_methods',
     'async',
     'description',
     'attach'
     ])

# Ports
PORT_DETAILED_RESOURCE = Resource(
    ['uuid',
     'address',
     'created_at',
     'extra',
     'node_uuid',
     'category',
     'vendor',
     'local_link_connection',
     'portgroup_uuid',
     'pxe_enabled',
     'physical_network',
     'updated_at',
     'internal_info',
     'is_smartnic',
     'name',
     ],
    sort_excluded=[
        'extra',
        # The server cannot sort on "node_uuid" or "portgroup_uuid" because
        # they aren't columns in the "ports" database table. "node_id" and
        # "portgroup_id" are stored, but it is internal to ironic.
        # See bug #1443003 for more details.
        'node_uuid',
        'portgroup_uuid',
        'internal_info',
    ])
PORT_RESOURCE = Resource(
    ['uuid',
     'address',
     ])

# Portgroups
PORTGROUP_DETAILED_RESOURCE = Resource(
    ['uuid',
     'address',
     'created_at',
     'extra',
     'standalone_ports_supported',
     'node_uuid',
     'name',
     'updated_at',
     'internal_info',
     'mode',
     'properties',
     'physical_network',
     'category',
     ],
    sort_excluded=[
        'extra',
        # The server cannot sort on "node_uuid" because it isn't a column in
        # the "portgroups" database table. "node_id" is stored, but it is
        # internal to ironic. See bug #1443003 for more details.
        'node_uuid',
        'internal_info',
        'properties',
    ])
PORTGROUP_RESOURCE = Resource(
    ['uuid',
     'address',
     'name',
     ])

# VIFs
VIF_RESOURCE = Resource(
    ['id'],
)

TRAIT_RESOURCE = Resource(
    ['traits'],
)

CHILDREN_RESOURCE = Resource(
    ['children'],
)

BIOS_RESOURCE = Resource(
    ['name', 'value'],
    override_labels={'name': 'BIOS setting name',
                     'value': 'BIOS setting value'}
)

BIOS_DETAILED_RESOURCE = Resource(
    ['name',
     'value',
     'attribute_type',
     'allowable_values',
     'lower_bound',
     'min_length',
     'max_length',
     'read_only',
     'reset_required',
     'unique',
     'upper_bound'
     ],
)

# Drivers
DRIVER_DETAILED_RESOURCE = Resource(
    ['name',
     'type',
     'hosts',
     'default_bios_interface',
     'default_boot_interface',
     'default_console_interface',
     'default_deploy_interface',
     'default_firmware_interface',
     'default_inspect_interface',
     'default_management_interface',
     'default_network_interface',
     'default_power_interface',
     'default_raid_interface',
     'default_rescue_interface',
     'default_storage_interface',
     'default_vendor_interface',
     'enabled_bios_interfaces',
     'enabled_boot_interfaces',
     'enabled_console_interfaces',
     'enabled_deploy_interfaces',
     'enabled_firmware_interfaces',
     'enabled_inspect_interfaces',
     'enabled_management_interfaces',
     'enabled_network_interfaces',
     'enabled_power_interfaces',
     'enabled_raid_interfaces',
     'enabled_rescue_interfaces',
     'enabled_storage_interfaces',
     'enabled_vendor_interfaces'
     ],
    override_labels={'name': 'Supported driver(s)'}
)
DRIVER_RESOURCE = Resource(
    ['name',
     'hosts',
     ],
    override_labels={'name': 'Supported driver(s)'}
)

# Firmware
FIRMWARE_RESOURCE = Resource(
    ['component',
     'initial_version',
     'current_version',
     'last_version_flashed',
     'created_at',
     'updated_at',
     ],
    override_labels={'component': 'Component',
                     'initial_version': 'Initial Version',
                     'current_version': 'Current Version',
                     'last_version_flashed': 'Last Version Flashed'}
)


# Volume connectors
VOLUME_CONNECTOR_DETAILED_RESOURCE = Resource(
    ['uuid',
     'node_uuid',
     'type',
     'connector_id',
     'extra',
     'created_at',
     'updated_at',
     ],
    sort_excluded=[
        # The server cannot sort on "node_uuid" because it isn't a column in
        # the "volume_connectors" database table. "node_id" is stored, but it
        # is internal to ironic. See bug #1443003 for more details.
        'node_uuid',
        'extra',
    ])
VOLUME_CONNECTOR_RESOURCE = Resource(
    ['uuid',
     'node_uuid',
     'type',
     'connector_id',
     ],
    sort_excluded=['node_uuid']
)

# Volume targets
VOLUME_TARGET_DETAILED_RESOURCE = Resource(
    ['uuid',
     'node_uuid',
     'volume_type',
     'properties',
     'boot_index',
     'extra',
     'volume_id',
     'created_at',
     'updated_at',
     ],
    sort_excluded=[
        # The server cannot sort on "node_uuid" because it isn't a column in
        # the "volume_targets" database table. "node_id" is stored, but it
        # is internal to ironic. See bug #1443003 for more details.
        'node_uuid',
        'extra',
        'properties'
    ])
VOLUME_TARGET_RESOURCE = Resource(
    ['uuid',
     'node_uuid',
     'volume_type',
     'boot_index',
     'volume_id',
     ],
    sort_excluded=['node_uuid']
)

# Conductors
CONDUCTOR_DETAILED_RESOURCE = Resource(
    ['hostname',
     'conductor_group',
     'alive',
     'drivers',
     'created_at',
     'updated_at',
     ],
    sort_excluded=[
        'alive',
        'drivers',
    ])
CONDUCTOR_RESOURCE = Resource(
    ['hostname',
     'conductor_group',
     'alive',
     ],
    sort_excluded=['alive']
)

# Allocations
ALLOCATION_DETAILED_RESOURCE = Resource(
    ['uuid',
     'name',
     'state',
     'owner',
     'node_uuid',
     'last_error',
     'resource_class',
     'traits',
     'candidate_nodes',
     'extra',
     'created_at',
     'updated_at',
     ],
    sort_excluded=[
        'candidate_nodes',
        'traits',
    ])
ALLOCATION_RESOURCE = Resource(
    ['uuid',
     'name',
     'resource_class',
     'state',
     'node_uuid',
     ],
)

# Deploy templates
DEPLOY_TEMPLATE_DETAILED_RESOURCE = Resource(
    ['uuid',
     'name',
     'steps',
     'extra',
     'created_at',
     'updated_at',
     ],
    sort_excluded=['extra', 'steps']
)

DEPLOY_TEMPLATE_RESOURCE = Resource(
    ['uuid',
     'name',
     ],
)

# Runbooks
RUNBOOK_DETAILED_RESOURCE = Resource(
    ['uuid',
     'name',
     'description',
     'traits',
     'owner',
     'public',
     'steps',
     'extra',
     'created_at',
     'updated_at'
     ],
    sort_excluded=['extra', 'steps', 'traits']
)

RUNBOOK_RESOURCE = Resource(
    ['uuid',
     'name',
     ],
)

NODE_HISTORY_RESOURCE = Resource(
    ['uuid',
     'created_at',
     'severity',
     'event']
)

NODE_HISTORY_DETAILED_RESOURCE = Resource(
    ['uuid',
     'created_at',
     'severity',
     'event_type',
     'event',
     'conductor',
     'user']
)

SHARD_RESOURCE = Resource(
    ['name',
     'count']
)

INSPECTION_RULE_DETAILED_RESOURCE = Resource(
    ['uuid',
     'description',
     'priority',
     'sensitive',
     'phase',
     'actions',
     'conditions',
     'created_at',
     'updated_at'],
    sort_excluded=['actions', 'conditions']
)

INSPECTION_RULE_RESOURCE = Resource(
    ['uuid',
     'description'],
)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import copy
import json
from unittest import mock

from osc_lib.tests import utils as osctestutils

from ironicclient import exc
from ironicclient.osc.v1 import baremetal_runbook
from ironicclient.tests.unit.osc.v1 import fakes as baremetal_fakes


class TestBaremetalRunbook(baremetal_fakes.TestBaremetal):

    def setUp(self) -> None:
        super(TestBaremetalRunbook, self).setUp()

        self.baremetal_mock = self.app.client_manager.baremetal
        self.baremetal_mock.reset_mock()
        # Set API version to support trait operations
        self.baremetal_mock.current_api_version = "1.112"


class TestCreateBaremetalRunbook(TestBaremetalRunbook):
    def setUp(self) -> None:
        super(TestCreateBaremetalRunbook, self).setUp()

        self.baremetal_mock.runbook.create.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.RUNBOOK),
                loaded=True,
            ))

        # Get the command object to test
        self.cmd = baremetal_runbook.CreateBaremetalRunbook(
            self.app, None)

    def test_baremetal_runbook_create(self) -> None:
        arglist = [
            '--name', baremetal_fakes.baremetal_runbook_name,
            '--steps', baremetal_fakes.baremetal_runbook_steps,
        ]

        verifylist = [
            ('name', baremetal_fakes.baremetal_runbook_name),
            ('steps', baremetal_fakes.baremetal_runbook_steps),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # Set expected values
        args = {
            'name': baremetal_fakes.baremetal_runbook_name,
            'steps': json.loads(
                baremetal_fakes.baremetal_runbook_steps),
        }

        self.baremetal_mock.runbook.create.assert_called_once_with(
            **args)

    def test_baremetal_runbook_create_uuid(self) -> None:
        arglist = [
            '--name', baremetal_fakes.baremetal_runbook_name,
            '--steps', baremetal_fakes.baremetal_runbook_steps,
            '--uuid', baremetal_fakes.baremetal_runbook_uuid,
        ]

        verifylist = [
            ('name', baremetal_fakes.baremetal_runbook_name),
            ('steps', baremetal_fakes.baremetal_runbook_steps),
            ('uuid', baremetal_fakes.baremetal_runbook_uuid),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # Set expected values
        args = {
            'name': baremetal_fakes.baremetal_runbook_name,
            'steps': json.loads(
                baremetal_fakes.baremetal_runbook_steps),
            'uuid': baremetal_fakes.baremetal_runbook_uuid,
        }

        self.baremetal_mock.runbook.create.assert_called_once_with(
            **args)

    def test_baremetal_runbook_create_no_name(self) -> None:
        arglist = [
            '--steps', baremetal_fakes.baremetal_runbook_steps,
        ]

        verifylist = [
            ('steps', baremetal_fakes.baremetal_runbook_steps),
        ]

        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)
        self.assertFalse(self.baremetal_mock.runbook.create.called)

    def test_baremetal_runbook_create_no_steps(self) -> None:
        arglist = [
            '--name', baremetal_fakes.baremetal_runbook_name,
        ]

        verifylist = [
            ('name', baremetal_fakes.baremetal_runbook_name),
        ]

        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)
        self.assertFalse(self.baremetal_mock.runbook.create.called)

    def test_baremetal_runbook_create_public_false(self) -> None:
        arglist = [
            '--name', baremetal_fakes.baremetal_runbook_name,
            '--steps', baremetal_fakes.baremetal_runbook_steps,
            '--public', 'false',
        ]
        verifylist = [
            ('name', baremetal_fakes.baremetal_runbook_name),
            ('steps', baremetal_fakes.baremetal_runbook_steps),
            ('public', 'false'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        expected = {
            'name': baremetal_fakes.baremetal_runbook_name,
            'steps': json.loads(baremetal_fakes.baremetal_runbook_steps),
            'public': False,
        }
        self.baremetal_mock.runbook.create.assert_called_once_with(**expected)


class TestShowBaremetalRunbook(TestBaremetalRunbook):
    def setUp(self) -> None:
        super(TestShowBaremetalRunbook, self).setUp()

        self.baremetal_mock.runbook.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.RUNBOOK),
                loaded=True))

        self.cmd = baremetal_runbook.ShowBaremetalRunbook(
            self.app, None)

    def test_baremetal_runbook_show(self) -> None:
        arglist = [baremetal_fakes.baremetal_runbook_uuid]
        verifylist = [('runbook',
                      baremetal_fakes.baremetal_runbook_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        args = [baremetal_fakes.baremetal_runbook_uuid]
        self.baremetal_mock.runbook.get.assert_called_with(
            *args, fields=None)

        collist = (
            'description',
            'extra',
            'name',
            'owner',
            'public',
            'steps',
            'traits',
            'uuid')
        self.assertEqual(collist, columns)

        datalist = (
            baremetal_fakes.baremetal_runbook_description,
            baremetal_fakes.baremetal_runbook_extra,
            baremetal_fakes.baremetal_runbook_name,
            baremetal_fakes.baremetal_runbook_owner,
            baremetal_fakes.baremetal_runbook_public,
            baremetal_fakes.baremetal_runbook_steps,
            baremetal_fakes.baremetal_runbook_traits,
            baremetal_fakes.baremetal_runbook_uuid)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_runbook_show_no_template(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestBaremetalRunbookSet(TestBaremetalRunbook):
    def setUp(self) -> None:
        super(TestBaremetalRunbookSet, self).setUp()

        self.baremetal_mock.runbook.update.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.RUNBOOK),
                loaded=True))

        self.cmd = baremetal_runbook.SetBaremetalRunbook(
            self.app, None)

    def test_baremetal_runbook_set_name(self) -> None:
        new_name = 'foo'
        arglist = [
            baremetal_fakes.baremetal_runbook_uuid,
            '--name', new_name]
        verifylist = [
            ('runbook', baremetal_fakes.baremetal_runbook_uuid),
            ('name', new_name)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.runbook.update.assert_called_once_with(
            baremetal_fakes.baremetal_runbook_uuid,
            [{'path': '/name', 'value': new_name, 'op': 'add'}])

    def test_baremetal_runbook_set_steps(self) -> None:
        arglist = [
            baremetal_fakes.baremetal_runbook_uuid,
            '--steps', baremetal_fakes.baremetal_runbook_steps]
        verifylist = [
            ('runbook', baremetal_fakes.baremetal_runbook_uuid),
            ('steps', baremetal_fakes.baremetal_runbook_steps)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        expected_steps = json.loads(
            baremetal_fakes.baremetal_runbook_steps)
        self.cmd.take_action(parsed_args)
        self.baremetal_mock.runbook.update.assert_called_once_with(
            baremetal_fakes.baremetal_runbook_uuid,
            [{'path': '/steps', 'value': expected_steps, 'op': 'add'}])

    def test_baremetal_runbook_set_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestBaremetalRunbookUnset(TestBaremetalRunbook):
    def setUp(self) -> None:
        super(TestBaremetalRunbookUnset, self).setUp()

        self.baremetal_mock.runbook.update.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.RUNBOOK),
                loaded=True))

        self.cmd = baremetal_runbook.UnsetBaremetalRunbook(
            self.app, None)

    def test_baremetal_runbook_unset_extra(self) -> None:
        arglist = [
            baremetal_fakes.baremetal_runbook_uuid, '--extra', 'key1']
        verifylist = [('runbook',
                       baremetal_fakes.baremetal_runbook_uuid),
                      ('extra', ['key1'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.runbook.update.assert_called_once_with(
            baremetal_fakes.baremetal_runbook_uuid,
            [{'path': '/extra/key1', 'op': 'remove'}])

    def test_baremetal_runbook_unset_multiple_extras(self) -> None:
        arglist = [
            baremetal_fakes.baremetal_runbook_uuid,
            '--extra', 'key1', '--extra', 'key2']
        verifylist = [('runbook',
                       baremetal_fakes.baremetal_runbook_uuid),
                      ('extra', ['key1', 'key2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.runbook.update.assert_called_once_with(
            baremetal_fakes.baremetal_runbook_uuid,
            [{'path': '/extra/key1', 'op': 'remove'},
             {'path': '/extra/key2', 'op': 'remove'}])

    def test_baremetal_runbook_unset_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_runbook_unset_no_property(self) -> None:
        uuid = baremetal_fakes.baremetal_runbook_uuid
        arglist = [uuid]
        verifylist = [('runbook', uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.assertFalse(self.baremetal_mock.runbook.update.called)


class TestBaremetalRunbookDelete(TestBaremetalRunbook):
    def setUp(self) -> None:
        super(TestBaremetalRunbookDelete, self).setUp()

        self.cmd = baremetal_runbook.DeleteBaremetalRunbook(
            self.app, None)

    def test_baremetal_runbook_delete(self) -> None:
        arglist = ['zzz-zzzzzz-zzzz']
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        args = 'zzz-zzzzzz-zzzz'
        self.baremetal_mock.runbook.delete.assert_called_with(args)

    def test_baremetal_runbook_delete_multiple(self) -> None:
        arglist = ['zzz-zzzzzz-zzzz', 'fakename']
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        args = ['zzz-zzzzzz-zzzz', 'fakename']
        self.baremetal_mock.runbook.delete.assert_has_calls(
            [mock.call(x) for x in args])
        self.assertEqual(
            2, self.baremetal_mock.runbook.delete.call_count)

    def test_baremetal_runbook_delete_multiple_with_fail(self) -> None:
        arglist = ['zzz-zzzzzz-zzzz', 'badname']
        verifylist = []

        self.baremetal_mock.runbook.delete.side_effect = [
            '', exc.ClientException]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(exc.ClientException,
                          self.cmd.take_action,
                          parsed_args)

        args = ['zzz-zzzzzz-zzzz', 'badname']
        self.baremetal_mock.runbook.delete.assert_has_calls(
            [mock.call(x) for x in args])
        self.assertEqual(
            2, self.baremetal_mock.runbook.delete.call_count)

    def test_baremetal_runbook_delete_no_template(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestBaremetalRunbookList(TestBaremetalRunbook):
    def setUp(self) -> None:
        super(TestBaremetalRunbookList, self).setUp()

        self.baremetal_mock.runbook.list.return_value = [
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.RUNBOOK),
                loaded=True)
        ]

        self.cmd = baremetal_runbook.ListBaremetalRunbook(
            self.app, None)

    def test_baremetal_runbook_list(self) -> None:
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None}
        self.baremetal_mock.runbook.list.assert_called_with(**kwargs)

        collist = (
            "uuid",
            "name")
        self.assertEqual(collist, columns)

        datalist = ((
            baremetal_fakes.baremetal_runbook_uuid,
            baremetal_fakes.baremetal_runbook_name
        ), )
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_runbook_list_long(self) -> None:
        arglist = ['--long']
        verifylist = [('detail', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'detail': True,
            'marker': None,
            'limit': None,
        }
        self.baremetal_mock.runbook.list.assert_called_with(**kwargs)

        collist = ('uuid',
                   'name',
                   'description',
                   'traits',
                   'owner',
                   'public',
                   'steps',
                   'extra',
                   'created_at',
                   'updated_at')
        self.assertEqual(collist, columns)

        datalist = ((
            baremetal_fakes.baremetal_runbook_uuid,
            baremetal_fakes.baremetal_runbook_name,
            baremetal_fakes.baremetal_runbook_description,
            baremetal_fakes.baremetal_runbook_traits,
            baremetal_fakes.baremetal_runbook_owner,
            baremetal_fakes.baremetal_runbook_public,
            baremetal_fakes.baremetal_runbook_steps,
            baremetal_fakes.baremetal_runbook_extra,
            '',
            '',
        ), )
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_runbook_list_fields(self) -> None:
        arglist = ['--fields', 'uuid', 'steps']
        verifylist = [('fields', [['uuid', 'steps']])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None,
            'detail': False,
            'fields': ('uuid', 'steps')
        }
        self.baremetal_mock.runbook.list.assert_called_with(**kwargs)

    def test_baremetal_runbook_list_fields_multiple(self) -> None:
        arglist = ['--fields', 'uuid', 'name', '--fields', 'steps']
        verifylist = [('fields', [['uuid', 'name'], ['steps']])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None,
            'detail': False,
            'fields': ('uuid', 'name', 'steps')
        }
        self.baremetal_mock.runbook.list.assert_called_with(**kwargs)

    def test_baremetal_runbook_list_invalid_fields(self) -> None:
        arglist = ['--fields', 'uuid', 'invalid']
        verifylist = [('fields', [['uuid', 'invalid']])]
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestListTraitsRunbook(TestBaremetalRunbook):
    def setUp(self) -> None:
        super(TestListTraitsRunbook, self).setUp()

        self.baremetal_mock.runbook.get_traits.return_value = (
            baremetal_fakes.TRAITS)

        # Get the command object to test
        self.cmd = baremetal_runbook.ListTraitsBaremetalRunbook(self.app, None)

    def test_baremetal_list_runbook_traits(self) -> None:
        arglist = ['runbook_uuid']
        verifylist = [('runbook', 'runbook_uuid')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.baremetal_mock.runbook.get_traits.assert_called_once_with(
            'runbook_uuid')
        self.assertEqual(('traits',), columns)
        self.assertEqual([[trait] for trait in baremetal_fakes.TRAITS], data)


class TestAddTraitRunbook(TestBaremetalRunbook):
    def setUp(self) -> None:
        super(TestAddTraitRunbook, self).setUp()

        # Get the command object to test
        self.cmd = baremetal_runbook.AddTraitBaremetalRunbook(self.app, None)

    def test_baremetal_runbook_add_trait(self) -> None:
        arglist = ['runbook_uuid', 'CUSTOM_FOO']
        verifylist = [('runbook', 'runbook_uuid'), ('traits', ['CUSTOM_FOO'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.baremetal_mock.runbook.add_trait.assert_called_once_with(
            'runbook_uuid', 'CUSTOM_FOO')

    def test_baremetal_runbook_add_multiple_traits(self) -> None:
        arglist = ['runbook_uuid', 'CUSTOM_FOO', 'CUSTOM_BAR']
        verifylist = [('runbook', 'runbook_uuid'),
                      ('traits', ['CUSTOM_FOO', 'CUSTOM_BAR'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        calls = [
            mock.call('runbook_uuid', 'CUSTOM_FOO'),
            mock.call('runbook_uuid', 'CUSTOM_BAR'),
        ]
        self.baremetal_mock.runbook.add_trait.assert_has_calls(calls)


class TestRemoveTraitRunbook(TestBaremetalRunbook):
    def setUp(self) -> None:
        super(TestRemoveTraitRunbook, self).setUp()

        # Get the command object to test
        self.cmd = baremetal_runbook.RemoveTraitBaremetalRunbook(
            self.app, None)

    def test_baremetal_runbook_remove_trait(self) -> None:
        arglist = ['runbook_uuid', 'CUSTOM_FOO']
        verifylist = [('runbook', 'runbook_uuid'), ('traits', ['CUSTOM_FOO'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.baremetal_mock.runbook.remove_trait.assert_called_once_with(
            'runbook_uuid', 'CUSTOM_FOO')

    def test_baremetal_runbook_remove_multiple_traits(self) -> None:
        arglist = ['runbook_uuid', 'CUSTOM_FOO', 'CUSTOM_BAR']
        verifylist = [('runbook', 'runbook_uuid'),
                      ('traits', ['CUSTOM_FOO', 'CUSTOM_BAR'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        calls = [
            mock.call('runbook_uuid', 'CUSTOM_FOO'),
            mock.call('runbook_uuid', 'CUSTOM_BAR'),
        ]
        self.baremetal_mock.runbook.remove_trait.assert_has_calls(calls)

    def test_baremetal_runbook_remove_all_traits(self) -> None:
        arglist = ['--all', 'runbook_uuid']
        verifylist = [('remove_all', True), ('runbook', 'runbook_uuid')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.baremetal_mock.runbook.remove_all_traits.assert_called_once_with(
            'runbook_uuid')



# Copyright 2016 Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import copy

import testtools

from ironicclient import exc
from ironicclient.tests.unit import utils
import ironicclient.v1.port
import ironicclient.v1.volume_target

NODE_UUID = '55555555-4444-3333-2222-111111111111'
TARGET1 = {'uuid': '11111111-2222-3333-4444-555555555555',
           'node_uuid': NODE_UUID,
           'volume_type': 'iscsi',
           'properties': {'target_iqn': 'iqn.foo'},
           'boot_index': 0,
           'volume_id': '12345678',
           'extra': {}}

TARGET2 = {'uuid': '66666666-7777-8888-9999-000000000000',
           'node_uuid': NODE_UUID,
           'volume_type': 'fibre_channel',
           'properties': {'target_wwn': 'foobar'},
           'boot_index': 1,
           'volume_id': '87654321',
           'extra': {}}

CREATE_TARGET = copy.deepcopy(TARGET1)
del CREATE_TARGET['uuid']

CREATE_TARGET_WITH_UUID = copy.deepcopy(TARGET1)

UPDATED_TARGET = copy.deepcopy(TARGET1)
NEW_VALUE = '100'
UPDATED_TARGET['boot_index'] = NEW_VALUE

fake_responses = {
    '/v1/volume/targets':
    {
        'GET': (
            {},
            {"targets": [TARGET1]},
        ),
        'POST': (
            {},
            TARGET1
        ),
    },
    '/v1/volume/targets/?detail=True':
    {
        'GET': (
            {},
            {"targets": [TARGET1]},
        ),
    },
    '/v1/volume/targets/?fields=uuid,boot_index':
    {
        'GET': (
            {},
            {"targets": [TARGET1]},
        ),
    },
    '/v1/volume/targets/%s' % TARGET1['uuid']:
    {
        'GET': (
            {},
            TARGET1,
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            UPDATED_TARGET,
        ),
    },
    '/v1/volume/targets/%s?fields=uuid,boot_index' % TARGET1['uuid']:
    {
        'GET': (
            {},
            TARGET1,
        ),
    },
    '/v1/volume/targets/?detail=True&node=%s' % NODE_UUID:
    {
        'GET': (
            {},
            {"targets": [TARGET1]},
        ),
    },
    '/v1/volume/targets/?node=%s' % NODE_UUID:
    {
        'GET': (
            {},
            {"targets": [TARGET1]},
        ),
    }
}

fake_responses_pagination = {
    '/v1/volume/targets':
    {
        'GET': (
            {},
            {"targets": [TARGET1],
             "next": "http://127.0.0.1:6385/v1/volume/targets/?marker=%s" %
             TARGET1['uuid']}
        ),
    },
    '/v1/volume/targets/?limit=1':
    {
        'GET': (
            {},
            {"targets": [TARGET1],
             "next": "http://127.0.0.1:6385/v1/volume/targets/?limit=1"
                     "&marker=%s" % TARGET1['uuid']}
        ),
    },
    '/v1/volume/targets/?limit=1&marker=%s' % TARGET1['uuid']:
    {
        'GET': (
            {},
            {"targets": [TARGET2]}
        ),
    },
    '/v1/volume/targets/?marker=%s' % TARGET1['uuid']:
    {
        'GET': (
            {},
            {"targets": [TARGET2]}
        ),
    },
}

fake_responses_sorting = {
    '/v1/volume/targets/?sort_key=updated_at':
    {
        'GET': (
            {},
            {"targets": [TARGET2, TARGET1]}
        ),
    },
    '/v1/volume/targets/?sort_dir=desc':
    {
        'GET': (
            {},
            {"targets": [TARGET2, TARGET1]}
        ),
    },
}


class VolumeTargetManagerTestBase(testtools.TestCase):

    def _validate_obj(self, expect: dict[str, object],
                      obj: object) -> None:
        self.assertEqual(expect['uuid'], obj.uuid)
        self.assertEqual(expect['volume_type'], obj.volume_type)
        self.assertEqual(expect['boot_index'], obj.boot_index)
        self.assertEqual(expect['volume_id'], obj.volume_id)
        self.assertEqual(expect['node_uuid'], obj.node_uuid)

    def _validate_list(self, expect_request: list[object],
                       expect_targets: list[dict[str, object]],
                       actual_targets: list[object]) -> None:
        self.assertEqual(expect_request, self.api.calls)
        self.assertEqual(len(expect_targets), len(actual_targets))
        for expect, obj in zip(expect_targets, actual_targets):
            self._validate_obj(expect, obj)


class VolumeTargetManagerTest(VolumeTargetManagerTestBase):

    def setUp(self) -> None:
        super(VolumeTargetManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = ironicclient.v1.volume_target.VolumeTargetManager(self.api)

    def test_volume_targets_list(self) -> None:
        volume_targets = self.mgr.list()
        expect = [
            ('GET', '/v1/volume/targets', {}, None),
        ]
        expect_targets = [TARGET1]
        self._validate_list(expect, expect_targets, volume_targets)

    def test_volume_targets_list_by_node(self) -> None:
        volume_targets = self.mgr.list(node=NODE_UUID)
        expect = [
            ('GET', '/v1/volume/targets/?node=%s' % NODE_UUID, {}, None),
        ]
        expect_targets = [TARGET1]
        self._validate_list(expect, expect_targets, volume_targets)

    def test_volume_targets_list_by_node_detail(self) -> None:
        volume_targets = self.mgr.list(node=NODE_UUID, detail=True)
        expect = [
            ('GET', '/v1/volume/targets/?detail=True&node=%s' % NODE_UUID,
             {}, None),
        ]
        expect_targets = [TARGET1]
        self._validate_list(expect, expect_targets, volume_targets)

    def test_volume_targets_list_detail(self) -> None:
        volume_targets = self.mgr.list(detail=True)
        expect = [
            ('GET', '/v1/volume/targets/?detail=True', {}, None),
        ]
        expect_targets = [TARGET1]
        self._validate_list(expect, expect_targets, volume_targets)

    def test_volume_target_list_fields(self) -> None:
        volume_targets = self.mgr.list(fields=['uuid', 'boot_index'])
        expect = [
            ('GET', '/v1/volume/targets/?fields=uuid,boot_index', {}, None),
        ]
        expect_targets = [TARGET1]
        self._validate_list(expect, expect_targets, volume_targets)

    def test_volume_target_list_detail_and_fields_fail(self) -> None:
        self.assertRaises(exc.InvalidAttribute, self.mgr.list,
                          detail=True, fields=['uuid', 'boot_index'])

    def test_volume_targets_show(self) -> None:
        volume_target = self.mgr.get(TARGET1['uuid'])
        expect = [
            ('GET', '/v1/volume/targets/%s' % TARGET1['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self._validate_obj(TARGET1, volume_target)

    def test_volume_target_show_fields(self) -> None:
        volume_target = self.mgr.get(TARGET1['uuid'],
                                     fields=['uuid', 'boot_index'])
        expect = [
            ('GET', '/v1/volume/targets/%s?fields=uuid,boot_index' %
             TARGET1['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(TARGET1['uuid'], volume_target.uuid)
        self.assertEqual(TARGET1['boot_index'], volume_target.boot_index)

    def test_create(self) -> None:
        volume_target = self.mgr.create(**CREATE_TARGET)
        expect = [
            ('POST', '/v1/volume/targets', {}, CREATE_TARGET),
        ]
        self.assertEqual(expect, self.api.calls)
        self._validate_obj(TARGET1, volume_target)

    def test_create_with_uuid(self) -> None:
        volume_target = self.mgr.create(**CREATE_TARGET_WITH_UUID)
        expect = [
            ('POST', '/v1/volume/targets', {}, CREATE_TARGET_WITH_UUID),
        ]
        self.assertEqual(expect, self.api.calls)
        self._validate_obj(TARGET1, volume_target)

    def test_delete(self) -> None:
        self.mgr.delete(TARGET1['uuid'])
        expect = [
            ('DELETE', '/v1/volume/targets/%s' % TARGET1['uuid'],
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)

    def test_update(self) -> None:
        patch = {'op': 'replace',
                 'value': NEW_VALUE,
                 'path': '/boot_index'}
        volume_target = self.mgr.update(
            volume_target_id=TARGET1['uuid'], patch=patch)
        expect = [
            ('PATCH', '/v1/volume/targets/%s' % TARGET1['uuid'],
             {}, patch),
        ]
        self.assertEqual(expect, self.api.calls)
        self._validate_obj(UPDATED_TARGET, volume_target)


class VolumeTargetManagerPaginationTest(VolumeTargetManagerTestBase):

    def setUp(self) -> None:
        super(VolumeTargetManagerPaginationTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.volume_target.VolumeTargetManager(self.api)

    def test_volume_targets_list_limit(self) -> None:
        volume_targets = self.mgr.list(limit=1)
        expect = [
            ('GET', '/v1/volume/targets/?limit=1', {}, None),
        ]
        expect_targets = [TARGET1]
        self._validate_list(expect, expect_targets, volume_targets)

    def test_volume_targets_list_marker(self) -> None:
        volume_targets = self.mgr.list(marker=TARGET1['uuid'])
        expect = [
            ('GET', '/v1/volume/targets/?marker=%s' % TARGET1['uuid'],
             {}, None),
        ]
        expect_targets = [TARGET2]
        self._validate_list(expect, expect_targets, volume_targets)

    def test_volume_targets_list_pagination_no_limit(self) -> None:
        volume_targets = self.mgr.list(limit=0)
        expect = [
            ('GET', '/v1/volume/targets', {}, None),
            ('GET', '/v1/volume/targets/?marker=%s' % TARGET1['uuid'],
             {}, None)
        ]
        expect_targets = [TARGET1, TARGET2]
        self._validate_list(expect, expect_targets, volume_targets)


class VolumeTargetManagerSortingTest(VolumeTargetManagerTestBase):

    def setUp(self) -> None:
        super(VolumeTargetManagerSortingTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = ironicclient.v1.volume_target.VolumeTargetManager(self.api)

    def test_volume_targets_list_sort_key(self) -> None:
        volume_targets = self.mgr.list(sort_key='updated_at')
        expect = [
            ('GET', '/v1/volume/targets/?sort_key=updated_at', {}, None)
        ]
        expect_targets = [TARGET2, TARGET1]
        self._validate_list(expect, expect_targets, volume_targets)

    def test_volume_targets_list_sort_dir(self) -> None:
        volume_targets = self.mgr.list(sort_dir='desc')
        expect = [
            ('GET', '/v1/volume/targets/?sort_dir=desc', {}, None)
        ]
        expect_targets = [TARGET2, TARGET1]
        self._validate_list(expect, expect_targets, volume_targets)



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import base64
import contextlib
import gzip
import json
import os
import shutil
import subprocess
import sys
import tempfile
import time
from typing import Any, Callable, Generator, Iterator, Protocol

from cliff import columns
from oslo_utils import strutils
import yaml

from ironicclient.common.http import _Version
from ironicclient.common.i18n import _
from ironicclient import exc


class ListCommandArgs(Protocol):
    """Protocol for parsed args passed to list commands (marker, limit, etc).

    """

    marker: str | None
    limit: int | None
    sort_key: str | None
    sort_dir: str | None
    detail: bool
    fields: list[list[str]] | None


def split_and_deserialize(string: str) -> tuple[str, Any]:
    """Split and try to JSON deserialize a string.

    Gets a string with the KEY=VALUE format, split it (using '=' as the
    separator) and try to JSON deserialize the VALUE.

    :returns: A tuple of (key, value).
    """
    try:
        key, value = string.split("=", 1)
    except ValueError:
        raise exc.CommandError(_('Attributes must be a list of '
                                 'PATH=VALUE not "%s"') % string)
    try:
        value = json.loads(value)
    except ValueError:
        pass

    return (key, value)


def key_value_pairs_to_dict(
    key_value_pairs: list[str] | None,
) -> dict[str, Any]:
    """Convert a list of key-value pairs to a dictionary.

    :param key_value_pairs: a list of strings, each string is in the form
                            <key>=<value>
    :returns: a dictionary, possibly empty
    """
    if key_value_pairs:
        return dict(split_and_deserialize(v) for v in key_value_pairs)
    return {}


def args_array_to_dict(
    kwargs: dict[str, Any],
    key_to_convert: str,
) -> dict[str, Any]:
    """Convert the value in a dictionary entry to a dictionary.

    From the kwargs dictionary, converts the value of the key_to_convert
    entry from a list of key-value pairs to a dictionary.

    :param kwargs: a dictionary
    :param key_to_convert: the key (in kwargs), whose value is expected to
        be a list of key=value strings. This value will be converted to a
        dictionary.
    :returns: kwargs, the (modified) dictionary
    """
    values_to_convert = kwargs.get(key_to_convert)
    if values_to_convert:
        kwargs[key_to_convert] = key_value_pairs_to_dict(values_to_convert)
    return kwargs


def args_array_to_patch(
    op: str,
    attributes: list[str],
) -> list[dict[str, Any]]:
    patch = []
    for attr in attributes:
        # Sanitize
        if not attr.startswith('/'):
            attr = '/' + attr

        if op in ['add', 'replace']:
            path, value = split_and_deserialize(attr)
            patch.append({'op': op, 'path': path, 'value': value})

        elif op == "remove":
            # For remove only the key is needed
            patch.append({'op': op, 'path': attr})
        else:
            raise exc.CommandError(_('Unknown PATCH operation: %s') % op)
    return patch


def convert_list_props_to_comma_separated(
    data: dict[str, Any],
    props: list[str] | None = None,
) -> dict[str, Any]:
    """Convert the list-type properties to comma-separated strings

    :param data: the input dict object.
    :param props: the properties whose values will be converted.
        Default to None to convert all list-type properties of the input.
    :returns: the result dict instance.
    """
    result = dict(data)

    if props is None:
        props = list(data.keys())

    for prop in props:
        val = data.get(prop, None)
        if isinstance(val, list):
            result[prop] = ', '.join(map(str, val))

    return result


def common_params_for_list(
    args: ListCommandArgs,
    fields: list[str],
    field_labels: list[str],
) -> dict[str, Any]:
    """Generate 'params' dict that is common for every 'list' command.

    :param args: arguments from command line.
    :param fields: possible fields for sorting.
    :param field_labels: possible field labels for sorting.
    :returns: a dict with params to pass to the client method.
    """
    params: dict[str, Any] = {}
    if args.marker is not None:
        params['marker'] = args.marker
    if args.limit is not None:
        if args.limit < 0:
            raise exc.CommandError(
                _('Expected non-negative --limit, got %s') % args.limit)
        params['limit'] = args.limit

    if args.sort_key is not None:
        # Support using both heading and field name for sort_key
        fields_map = dict(zip(field_labels, fields))
        fields_map.update(zip(fields, fields))
        try:
            sort_key = fields_map[args.sort_key]
        except KeyError:
            raise exc.CommandError(
                _("%(sort_key)s is an invalid field for sorting, "
                  "valid values for --sort-key are: %(valid)s") %
                {'sort_key': args.sort_key,
                 'valid': list(fields_map)})
        params['sort_key'] = sort_key
    if args.sort_dir is not None:
        if args.sort_dir not in ('asc', 'desc'):
            raise exc.CommandError(
                _("%s is an invalid value for sort direction, "
                  "valid values for --sort-dir are: 'asc', 'desc'") %
                args.sort_dir)
        params['sort_dir'] = args.sort_dir

    params['detail'] = args.detail

    requested_fields = args.fields[0] if args.fields else None
    if requested_fields is not None:
        params['fields'] = requested_fields

    return params


def common_filters(
    marker: str | None = None,
    limit: int | None = None,
    sort_key: str | None = None,
    sort_dir: str | None = None,
    fields: list[str] | None = None,
    detail: bool = False,
    project: str | None = None,
    public: bool | None = None,
) -> list[str]:
    """Generate common filters for any list request.

    :param marker: entity ID from which to start returning entities.
    :param limit: maximum number of entities to return.
    :param sort_key: field to use for sorting.
    :param sort_dir: direction of sorting: 'asc' or 'desc'.
    :param fields: a list with a specified set of fields of the resource
                   to be returned.
    :param detail: Boolean, True to return detailed information. This parameter
                   can be used for resources which accept 'detail' as a URL
                   parameter.
    :returns: list of string filters.
    """
    filters = []
    if isinstance(limit, int) and limit > 0:
        filters.append('limit=%s' % limit)
    if marker is not None:
        filters.append('marker=%s' % marker)
    if sort_key is not None:
        filters.append('sort_key=%s' % sort_key)
    if sort_dir is not None:
        filters.append('sort_dir=%s' % sort_dir)
    if project is not None:
        filters.append('project=%s' % project)
    if public is not None:
        filters.append('public=True')
    if fields is not None:
        filters.append('fields=%s' % ','.join(fields))
    if detail:
        filters.append('detail=True')
    return filters


@contextlib.contextmanager
def tempdir(
    suffix: str | None = None,
    prefix: str | None = None,
    dir: str | None = None,
) -> Generator[str, None, None]:
    dirname = tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dir)
    try:
        yield dirname
    finally:
        shutil.rmtree(dirname)


def make_configdrive(path: str) -> bytes:
    """Make the config drive file.

    :param path: The directory containing the config drive files.
    :returns: A gzipped and base64 encoded configdrive string.

    """
    # Make sure path it's readable
    if not os.access(path, os.R_OK):
        raise exc.CommandError(_('The directory "%s" is not readable') % path)

    with tempfile.NamedTemporaryFile() as tmpfile:
        with tempfile.NamedTemporaryFile() as tmpzipfile:
            publisher = 'ironicclient-configdrive 0.1'
            # NOTE(toabctl): Luckily, genisoimage, mkisofs and xorrisofs
            # understand the same parameters which are currently used.
            cmds = ['genisoimage', 'mkisofs', 'xorrisofs']
            error: OSError | None = None
            p: subprocess.Popen[bytes] | None = None
            for c in cmds:
                try:
                    p = subprocess.Popen([c, '-o', tmpfile.name,
                                          '-ldots', '-allow-lowercase',
                                          '-allow-multidot', '-l',
                                          '-publisher', publisher,
                                          '-quiet', '-J',
                                          '-r', '-V', 'config-2',
                                          path],
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
                except OSError as e:
                    error = e
                else:
                    error = None
                    break
            if error or p is None:
                raise exc.CommandError(
                    _('Error generating the config drive. Make sure the '
                      '"genisoimage", "mkisofs", or "xorriso" tool is '
                      'installed. Error: %s') % error)

            stdout, stderr = p.communicate()
            if p.returncode != 0:
                raise exc.CommandError(
                    _('Error generating the config drive.'
                      'Stdout: "%(stdout)s". Stderr: %(stderr)s') %
                    {'stdout': stdout, 'stderr': stderr})

            # Compress file
            tmpfile.seek(0)
            with gzip.GzipFile(fileobj=tmpzipfile, mode='wb') as gz_file:
                shutil.copyfileobj(tmpfile, gz_file)

            tmpzipfile.seek(0)
            return base64.b64encode(tmpzipfile.read())


def check_empty_arg(arg: str, arg_descriptor: str) -> None:
    if not arg.strip():
        raise exc.CommandError(_('%(arg)s cannot be empty or only have blank'
                                 ' spaces') % {'arg': arg_descriptor})


def bool_argument_value(
    arg_name: str,
    bool_str: str,
    strict: bool = True,
    default: bool = False,
) -> bool:
    """Returns the Boolean represented by bool_str.

    Returns the Boolean value for the argument named arg_name. The value is
    represented by the string bool_str. If the string is an invalid Boolean
    string: if strict is True, a CommandError exception is raised; otherwise
    the default value is returned.

    :param arg_name: The name of the argument
    :param bool_str: The string representing a Boolean value
    :param strict: Used if the string is invalid. If True, raises an exception.
        If False, returns the default value.
    :param default: The default value to return if the string is invalid
        and not strict
    :returns: the Boolean value represented by bool_str or the default value
        if bool_str is invalid and strict is False
    :raises CommandError: if bool_str is an invalid Boolean string

    """
    try:
        val = strutils.bool_from_string(bool_str, strict, default)
    except ValueError as e:
        raise exc.CommandError(_("argument %(arg)s: %(err)s.")
                               % {'arg': arg_name, 'err': e})
    return val


def check_for_invalid_fields(
    fields: list[str] | None,
    valid_fields: list[str],
) -> None:
    """Check for invalid fields.

    :param fields: A list of fields specified by the user.
    :param valid_fields: A list of valid fields.
    :raises CommandError: If invalid fields were specified by the user.
    """
    if not fields:
        return

    invalid_fields = set(fields) - set(valid_fields)
    if invalid_fields:
        raise exc.CommandError(
            _('Invalid field(s) requested: %(invalid)s. Valid fields '
              'are: %(valid)s.') % {'invalid': ', '.join(invalid_fields),
                                    'valid': ', '.join(valid_fields)})


def get_from_stdin(info_desc: str) -> str:
    """Read information from stdin.

    :param info_desc: A string description of the desired information
    :raises: InvalidAttribute if there was a problem reading from stdin
    :returns: the string that was read from stdin
    """
    try:
        info = sys.stdin.read().strip()
    except Exception as e:
        err = _("Cannot get %(desc)s from standard input. Error: %(err)s")
        raise exc.InvalidAttribute(err % {'desc': info_desc, 'err': e})
    return info


def handle_json_or_file_arg(
    json_arg: str,
) -> list[Any] | dict[str, Any]:
    """Attempts to read JSON argument from file or string.

    :param json_arg: May be a file name containing the YAML or JSON, or
        a JSON string.
    :returns: A list or dictionary parsed from JSON.
    :raises: InvalidAttribute if the argument cannot be parsed.
    """

    if os.path.isfile(json_arg):
        try:
            with open(json_arg, 'r') as f:
                # safe_load returns Any; we treat as list | dict
                return yaml.safe_load(f)  # type: ignore[no-any-return]
        except Exception as e:
            err = _("Cannot get JSON/YAML from file '%(file)s'. "
                    "Error: %(err)s") % {'err': e, 'file': json_arg}
            raise exc.InvalidAttribute(err)
    try:
        result: list[Any] | dict[str, Any] = json.loads(json_arg)
    except ValueError as e:
        err = (_("Value '%(string)s' is not a file and cannot be parsed "
                 "as JSON: '%(err)s'") % {'err': e, 'string': json_arg})
        raise exc.InvalidAttribute(err)

    return result


def poll(
    timeout: int | float,
    poll_interval: int | float,
    poll_delay_function: Callable[[int | float], Any] | None,
    timeout_message: str | Callable[[], str],
) -> Iterator[int]:
    if not isinstance(timeout, (int, float)) or timeout < 0:
        raise ValueError(_('Timeout must be a non-negative number'))

    threshold = time.time() + timeout
    poll_delay_function = (time.sleep if poll_delay_function is None
                           else poll_delay_function)
    if not callable(poll_delay_function):
        raise TypeError(_('poll_delay_function must be callable'))

    count = 0
    while not timeout or time.time() < threshold:
        yield count

        poll_delay_function(poll_interval)
        count += 1

    if callable(timeout_message):
        timeout_message = timeout_message()
    raise exc.StateTransitionTimeout(timeout_message)


def handle_json_arg(
    json_arg: str | None,
    info_desc: str,
) -> list[Any] | dict[str, Any] | None:
    """Read a JSON argument from stdin, file or string.

    :param json_arg: May be a file name containing the JSON, a JSON string, or
        '-' indicating that the argument should be read from standard input.
    :param info_desc: A string description of the desired information
    :returns: A list or dictionary parsed from JSON, or None if json_arg is
        empty (e.g. after reading from stdin with no input).
    :raises: InvalidAttribute if the argument cannot be parsed.
    """
    if json_arg == '-':
        json_arg = get_from_stdin(info_desc)
    if json_arg:
        return handle_json_or_file_arg(json_arg)
    return None


def get_json_data(
    data: bytes,
) -> dict[str, Any] | list[Any] | None:
    """Check if the binary data is JSON and parse it if so.

    :param data: Raw bytes to parse (e.g. response body).
    :returns: Parsed dict or list if valid JSON, otherwise None.
    """
    # We don't want to simply loads() a potentially large binary. Doing so,
    # in my testing, is orders of magnitude (!!) slower than this process.
    for idx in range(len(data)):
        char = data[idx:idx + 1]
        if char.isspace():
            continue
        if char != b'{' and char != b'[':
            return None  # not JSON, at least not JSON we care about
        break  # maybe JSON

    try:
        # loads returns Any; we treat as dict | list
        return json.loads(data)  # type: ignore[no-any-return]
    except ValueError:
        return None


def format_hash(
    data: dict[str, Any] | None,
    prefix: str = "",
) -> str | None:
    if data is None:
        return None

    output = ""
    for s in sorted(data):
        key_str = prefix + "." + s if prefix else s
        if isinstance(data[s], dict):
            # NOTE(dtroyer): Only append the separator chars here, quoting
            #                is completely handled in the terminal case.
            hashed = format_hash(data[s], prefix=key_str)
            output = output + (hashed or "") + ", "
        elif data[s] is not None:
            output = output + key_str + "='" + str(data[s]) + "', "
        else:
            output = output + key_str + "=, "
    return output[:-2]


class HashColumn(columns.FormattableColumn[Any]):
    # base returns Sequence; we return str | None
    def human_readable(self) -> str | None:  # type: ignore[override]
        return format_hash(self._value)


def check_api_version_support(
    current_version: str | list[str],
    required_version: str,
) -> bool:
    """Check if the current API version supports a required version.

    :param current_version: Current API version being used (e.g. "1.112")
    :param required_version: Minimum API version required (e.g. "1.112")
    :returns: True if the current version meets the requirement, False
        otherwise. Also returns True if the version format is invalid
        (let the server decide).
    """
    version = (current_version[0] if isinstance(current_version, list)
               else current_version)

    try:
        return _Version(version) >= _Version(required_version)
    except ValueError:
        # Invalid version format - let the server handle it
        return True
